import torch
from torch import Tensor, device as Device, dtype as DType, nn

import imaginairy.vendored.refiners.fluxion.layers as fl
from imaginairy.vendored.refiners.fluxion.context import Contexts
from imaginairy.vendored.refiners.foundationals.segment_anything.transformer import (
    SparseCrossDenseAttention,
    TwoWayTranformerLayer,
)


class EmbeddingsAggregator(fl.ContextModule):
    def __init__(self, num_output_mask: int = 3) -> None:
        super().__init__()
        self.num_mask_tokens = num_output_mask

    def forward(self, iou_mask_tokens: Tensor) -> Tensor:
        mask_decoder = self.ensure_parent
        mask_decoder_context = mask_decoder.use_context(context_name="mask_decoder")
        image_embedding = mask_decoder_context["image_embedding"]
        point_embedding = mask_decoder_context["point_embedding"]
        mask_embedding = mask_decoder_context["mask_embedding"]
        dense_positional_embedding = mask_decoder_context["dense_positional_embedding"]

        sparse_embedding = torch.cat(tensors=(iou_mask_tokens, point_embedding), dim=1)
        dense_embedding = (image_embedding + mask_embedding).flatten(start_dim=2).transpose(1, 2)
        if dense_positional_embedding.shape != dense_embedding.shape:
            dense_positional_embedding = dense_positional_embedding.flatten(start_dim=2).transpose(1, 2)

        mask_decoder_context.update(
            {
                "dense_embedding": dense_embedding,
                "dense_positional_embedding": dense_positional_embedding,
                "sparse_embedding": sparse_embedding,
            }
        )
        mask_decoder.set_context(context="mask_decoder", value=mask_decoder_context)

        return sparse_embedding


class Transformer(fl.Chain):
    pass


class Hypernetworks(fl.Concatenate):
    def __init__(
        self,
        embedding_dim: int = 256,
        num_layers: int = 3,
        num_mask_tokens: int = 3,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers
        self.num_mask_tokens = num_mask_tokens

        super().__init__(
            *[
                fl.Chain(
                    fl.Slicing(dim=1, start=i + 1, end=i + 2),
                    fl.MultiLinear(
                        input_dim=embedding_dim,
                        output_dim=embedding_dim // 8,
                        inner_dim=embedding_dim,
                        num_layers=num_layers,
                        device=device,
                        dtype=dtype,
                    ),
                )
                for i in range(num_mask_tokens + 1)
            ],
            dim=1,
        )


class DenseEmbeddingUpscaling(fl.Chain):
    def __init__(
        self,
        embedding_dim: int = 256,
        dense_embedding_side_dim: int = 64,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.dense_embedding_side_dim = dense_embedding_side_dim

        super().__init__(
            fl.UseContext(context="mask_decoder", key="dense_embedding"),
            fl.Transpose(dim0=1, dim1=2),
            fl.Reshape(embedding_dim, dense_embedding_side_dim, dense_embedding_side_dim),
            fl.ConvTranspose2d(
                in_channels=embedding_dim,
                out_channels=embedding_dim // 4,
                kernel_size=2,
                stride=2,
                device=device,
                dtype=dtype,
            ),
            fl.LayerNorm2d(channels=embedding_dim // 4, device=device, dtype=dtype),
            fl.GeLU(),
            fl.ConvTranspose2d(
                in_channels=embedding_dim // 4,
                out_channels=embedding_dim // 8,
                kernel_size=2,
                stride=2,
                device=device,
                dtype=dtype,
            ),
            fl.GeLU(),
            fl.Flatten(start_dim=2),
        )


class IOUMaskEncoder(fl.WeightedModule):
    def __init__(
        self,
        embedding_dim: int = 256,
        num_mask_tokens: int = 4,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.num_mask_tokens = num_mask_tokens
        # aka prompt tokens + output token (for IoU scores prediction)
        self.weight = nn.Parameter(data=torch.randn(num_mask_tokens + 1, embedding_dim, device=device, dtype=dtype))

    def forward(self) -> Tensor:
        return self.weight.unsqueeze(dim=0)


class MaskPrediction(fl.Chain):
    def __init__(
        self,
        embedding_dim: int,
        num_mask_tokens: int,
        num_layers: int = 3,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.embedding_dim = embedding_dim
        self.num_mask_tokens = num_mask_tokens
        self.num_layers = num_layers
        super().__init__(
            fl.Matmul(
                input=Hypernetworks(
                    embedding_dim=embedding_dim,
                    num_layers=num_layers,
                    num_mask_tokens=num_mask_tokens,
                    device=device,
                    dtype=dtype,
                ),
                other=DenseEmbeddingUpscaling(embedding_dim=embedding_dim, device=device, dtype=dtype),
            ),
            fl.Slicing(dim=1, start=1),
            fl.Reshape(num_mask_tokens, embedding_dim, embedding_dim),
        )


class IOUPrediction(fl.Chain):
    def __init__(
        self,
        embedding_dim: int,
        num_layers: int,
        num_mask_tokens: int,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers
        super().__init__(
            fl.Slicing(dim=1, start=0, end=1),
            fl.Squeeze(dim=0),
            fl.MultiLinear(
                input_dim=embedding_dim,
                output_dim=num_mask_tokens + 1,
                inner_dim=embedding_dim,
                num_layers=num_layers,
                device=device,
                dtype=dtype,
            ),
            fl.Slicing(dim=-1, start=1),
        )


class MaskDecoder(fl.Chain):
    def __init__(
        self,
        embedding_dim: int = 256,
        feed_forward_dim: int = 2048,
        num_layers: int = 2,
        num_output_mask: int = 3,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.num_mask_tokens = num_output_mask
        self.feed_forward_dim = feed_forward_dim
        self.num_layers = num_layers

        super().__init__(
            IOUMaskEncoder(
                embedding_dim=embedding_dim, num_mask_tokens=num_output_mask + 1, device=device, dtype=dtype
            ),
            EmbeddingsAggregator(num_output_mask=num_output_mask),
            Transformer(
                *(
                    TwoWayTranformerLayer(
                        embedding_dim=embedding_dim,
                        num_heads=8,
                        feed_forward_dim=feed_forward_dim,
                        use_residual_self_attention=i > 0,
                        device=device,
                        dtype=dtype,
                    )
                    for i in range(num_layers)
                ),
                SparseCrossDenseAttention(embedding_dim=embedding_dim, device=device, dtype=dtype),
                fl.LayerNorm(normalized_shape=embedding_dim, device=device, dtype=dtype),
            ),
            fl.Parallel(
                MaskPrediction(
                    embedding_dim=embedding_dim, num_mask_tokens=num_output_mask, device=device, dtype=dtype
                ),
                IOUPrediction(
                    embedding_dim=embedding_dim,
                    num_layers=3,
                    num_mask_tokens=num_output_mask,
                    device=device,
                    dtype=dtype,
                ),
            ),
        )

    def init_context(self) -> Contexts:
        return {
            "mask_decoder": {
                "image_embedding": None,
                "point_embedding": None,
                "mask_embedding": None,
                "dense_positional_embedding": None,
            }
        }

    def set_image_embedding(self, image_embedding: Tensor) -> None:
        mask_decoder_context = self.use_context(context_name="mask_decoder")
        mask_decoder_context["image_embedding"] = image_embedding

    def set_point_embedding(self, point_embedding: Tensor) -> None:
        mask_decoder_context = self.use_context(context_name="mask_decoder")
        mask_decoder_context["point_embedding"] = point_embedding

    def set_mask_embedding(self, mask_embedding: Tensor) -> None:
        mask_decoder_context = self.use_context(context_name="mask_decoder")
        mask_decoder_context["mask_embedding"] = mask_embedding

    def set_dense_positional_embedding(self, dense_positional_embedding: Tensor) -> None:
        mask_decoder_context = self.use_context(context_name="mask_decoder")
        mask_decoder_context["dense_positional_embedding"] = dense_positional_embedding



"""Classes for video generation models"""

from typing import List, Optional, Union

import torch as th
import torch.nn as nn
from einops import rearrange

from imaginairy.modules.sgm.diffusionmodules.openaimodel import (
    Downsample,
    ResBlock,
    SpatialVideoTransformer,
    Timestep,
    TimestepEmbedSequential,
    Upsample,
)
from imaginairy.modules.sgm.diffusionmodules.util import (
    conv_nd,
    linear,
    normalization,
    timestep_embedding,
    zero_module,
)
from imaginairy.utils import default

from .util import AlphaBlender

# import torch.nn.functional as F


class VideoResBlock(ResBlock):
    def __init__(
        self,
        channels: int,
        emb_channels: int,
        dropout: float,
        video_kernel_size: Union[int, List[int]] = 3,
        merge_strategy: str = "fixed",
        merge_factor: float = 0.5,
        out_channels: Optional[int] = None,
        use_conv: bool = False,
        use_scale_shift_norm: bool = False,
        dims: int = 2,
        use_checkpoint: bool = False,
        up: bool = False,
        down: bool = False,
    ):
        super().__init__(
            channels,
            emb_channels,
            dropout,
            out_channels=out_channels,
            use_conv=use_conv,
            use_scale_shift_norm=use_scale_shift_norm,
            dims=dims,
            use_checkpoint=use_checkpoint,
            up=up,
            down=down,
        )

        self.time_stack = ResBlock(
            default(out_channels, channels),
            emb_channels,
            dropout=dropout,
            dims=3,
            out_channels=default(out_channels, channels),
            use_scale_shift_norm=False,
            use_conv=False,
            up=False,
            down=False,
            kernel_size=video_kernel_size,
            use_checkpoint=use_checkpoint,
            exchange_temb_dims=True,
        )
        self.time_mixer = AlphaBlender(
            alpha=merge_factor,
            merge_strategy=merge_strategy,
            rearrange_pattern="b t -> b 1 t 1 1",
        )

    def forward(
        self,
        x: th.Tensor,
        emb: th.Tensor,
        num_video_frames: int,
        image_only_indicator: Optional[th.Tensor] = None,
    ) -> th.Tensor:
        x = super().forward(x, emb)

        x_mix = rearrange(x, "(b t) c h w -> b c t h w", t=num_video_frames)
        x = rearrange(x, "(b t) c h w -> b c t h w", t=num_video_frames)

        x = self.time_stack(
            x, rearrange(emb, "(b t) ... -> b t ...", t=num_video_frames)
        )
        x = self.time_mixer(
            x_spatial=x_mix, x_temporal=x, image_only_indicator=image_only_indicator
        )
        x = rearrange(x, "b c t h w -> (b t) c h w")
        return x


class VideoUNet(nn.Module):
    def __init__(
        self,
        in_channels: int,
        model_channels: int,
        out_channels: int,
        num_res_blocks: int,
        attention_resolutions: int,
        dropout: float = 0.0,
        channel_mult: List[int] = (1, 2, 4, 8),
        conv_resample: bool = True,
        dims: int = 2,
        num_classes: Optional[int] = None,
        use_checkpoint: bool = False,
        num_heads: int = -1,
        num_head_channels: int = -1,
        num_heads_upsample: int = -1,
        use_scale_shift_norm: bool = False,
        resblock_updown: bool = False,
        transformer_depth: Union[List[int], int] = 1,
        transformer_depth_middle: Optional[int] = None,
        context_dim: Optional[int] = None,
        time_downup: bool = False,
        time_context_dim: Optional[int] = None,
        extra_ff_mix_layer: bool = False,
        use_spatial_context: bool = False,
        merge_strategy: str = "fixed",
        merge_factor: float = 0.5,
        spatial_transformer_attn_type: str = "softmax",
        video_kernel_size: Union[int, List[int]] = 3,
        use_linear_in_transformer: bool = False,
        adm_in_channels: Optional[int] = None,
        disable_temporal_crossattention: bool = False,
        max_ddpm_temb_period: int = 10000,
    ):
        super().__init__()
        assert context_dim is not None

        if num_heads_upsample == -1:
            num_heads_upsample = num_heads

        if num_heads == -1:
            assert num_head_channels != -1

        if num_head_channels == -1:
            assert num_heads != -1

        self.in_channels = in_channels
        self.model_channels = model_channels
        self.out_channels = out_channels
        if isinstance(transformer_depth, int):
            transformer_depth = len(channel_mult) * [transformer_depth]
        transformer_depth_middle = default(
            transformer_depth_middle, transformer_depth[-1]
        )

        self.num_res_blocks = num_res_blocks
        self.attention_resolutions = attention_resolutions
        self.dropout = dropout
        self.channel_mult = channel_mult
        self.conv_resample = conv_resample
        self.num_classes = num_classes
        self.use_checkpoint = use_checkpoint
        self.num_heads = num_heads
        self.num_head_channels = num_head_channels
        self.num_heads_upsample = num_heads_upsample

        time_embed_dim = model_channels * 4
        self.time_embed = nn.Sequential(
            linear(model_channels, time_embed_dim),
            nn.SiLU(),
            linear(time_embed_dim, time_embed_dim),
        )

        if self.num_classes is not None:
            if isinstance(self.num_classes, int):
                self.label_emb = nn.Embedding(num_classes, time_embed_dim)
            elif self.num_classes == "continuous":
                print("setting up linear c_adm embedding layer")
                self.label_emb = nn.Linear(1, time_embed_dim)
            elif self.num_classes == "timestep":
                self.label_emb = nn.Sequential(
                    Timestep(model_channels),
                    nn.Sequential(
                        linear(model_channels, time_embed_dim),
                        nn.SiLU(),
                        linear(time_embed_dim, time_embed_dim),
                    ),
                )

            elif self.num_classes == "sequential":
                assert adm_in_channels is not None
                self.label_emb = nn.Sequential(
                    nn.Sequential(
                        linear(adm_in_channels, time_embed_dim),
                        nn.SiLU(),
                        linear(time_embed_dim, time_embed_dim),
                    )
                )
            else:
                raise ValueError()

        self.input_blocks = nn.ModuleList(
            [
                TimestepEmbedSequential(
                    conv_nd(dims, in_channels, model_channels, 3, padding=1)
                )
            ]
        )
        self._feature_size = model_channels
        input_block_chans = [model_channels]
        ch = model_channels
        ds = 1

        def get_attention_layer(
            ch,
            num_heads,
            dim_head,
            depth=1,
            context_dim=None,
            use_checkpoint=False,
            disabled_sa=False,
        ):
            return SpatialVideoTransformer(
                ch,
                num_heads,
                dim_head,
                depth=depth,
                context_dim=context_dim,
                time_context_dim=time_context_dim,
                dropout=dropout,
                ff_in=extra_ff_mix_layer,
                use_spatial_context=use_spatial_context,
                merge_strategy=merge_strategy,
                merge_factor=merge_factor,
                checkpoint=use_checkpoint,
                use_linear=use_linear_in_transformer,
                attn_mode=spatial_transformer_attn_type,
                disable_self_attn=disabled_sa,
                disable_temporal_crossattention=disable_temporal_crossattention,
                max_time_embed_period=max_ddpm_temb_period,
            )

        def get_resblock(
            merge_factor,
            merge_strategy,
            video_kernel_size,
            ch,
            time_embed_dim,
            dropout,
            out_ch,
            dims,
            use_checkpoint,
            use_scale_shift_norm,
            down=False,
            up=False,
        ):
            return VideoResBlock(
                merge_factor=merge_factor,
                merge_strategy=merge_strategy,
                video_kernel_size=video_kernel_size,
                channels=ch,
                emb_channels=time_embed_dim,
                dropout=dropout,
                out_channels=out_ch,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
                down=down,
                up=up,
            )

        for level, mult in enumerate(channel_mult):
            for _ in range(num_res_blocks):
                layers = [
                    get_resblock(
                        merge_factor=merge_factor,
                        merge_strategy=merge_strategy,
                        video_kernel_size=video_kernel_size,
                        ch=ch,
                        time_embed_dim=time_embed_dim,
                        dropout=dropout,
                        out_ch=mult * model_channels,
                        dims=dims,
                        use_checkpoint=use_checkpoint,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                ]
                ch = mult * model_channels
                if ds in attention_resolutions:
                    if num_head_channels == -1:
                        dim_head = ch // num_heads
                    else:
                        num_heads = ch // num_head_channels
                        dim_head = num_head_channels

                    layers.append(
                        get_attention_layer(
                            ch,
                            num_heads,
                            dim_head,
                            depth=transformer_depth[level],
                            context_dim=context_dim,
                            use_checkpoint=use_checkpoint,
                            disabled_sa=False,
                        )
                    )
                self.input_blocks.append(TimestepEmbedSequential(*layers))
                self._feature_size += ch
                input_block_chans.append(ch)
            if level != len(channel_mult) - 1:
                ds *= 2
                out_ch = ch
                self.input_blocks.append(
                    TimestepEmbedSequential(
                        get_resblock(
                            merge_factor=merge_factor,
                            merge_strategy=merge_strategy,
                            video_kernel_size=video_kernel_size,
                            ch=ch,
                            time_embed_dim=time_embed_dim,
                            dropout=dropout,
                            out_ch=out_ch,
                            dims=dims,
                            use_checkpoint=use_checkpoint,
                            use_scale_shift_norm=use_scale_shift_norm,
                            down=True,
                        )
                        if resblock_updown
                        else Downsample(
                            ch,
                            conv_resample,
                            dims=dims,
                            out_channels=out_ch,
                            third_down=time_downup,
                        )
                    )
                )
                ch = out_ch
                input_block_chans.append(ch)

                self._feature_size += ch

        if num_head_channels == -1:
            dim_head = ch // num_heads
        else:
            num_heads = ch // num_head_channels
            dim_head = num_head_channels

        self.middle_block = TimestepEmbedSequential(
            get_resblock(
                merge_factor=merge_factor,
                merge_strategy=merge_strategy,
                video_kernel_size=video_kernel_size,
                ch=ch,
                time_embed_dim=time_embed_dim,
                out_ch=None,
                dropout=dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
            get_attention_layer(
                ch,
                num_heads,
                dim_head,
                depth=transformer_depth_middle,
                context_dim=context_dim,
                use_checkpoint=use_checkpoint,
            ),
            get_resblock(
                merge_factor=merge_factor,
                merge_strategy=merge_strategy,
                video_kernel_size=video_kernel_size,
                ch=ch,
                out_ch=None,
                time_embed_dim=time_embed_dim,
                dropout=dropout,
                dims=dims,
                use_checkpoint=use_checkpoint,
                use_scale_shift_norm=use_scale_shift_norm,
            ),
        )
        self._feature_size += ch

        self.output_blocks = nn.ModuleList([])
        for level, mult in list(enumerate(channel_mult))[::-1]:
            for i in range(num_res_blocks + 1):
                ich = input_block_chans.pop()
                layers = [
                    get_resblock(
                        merge_factor=merge_factor,
                        merge_strategy=merge_strategy,
                        video_kernel_size=video_kernel_size,
                        ch=ch + ich,
                        time_embed_dim=time_embed_dim,
                        dropout=dropout,
                        out_ch=model_channels * mult,
                        dims=dims,
                        use_checkpoint=use_checkpoint,
                        use_scale_shift_norm=use_scale_shift_norm,
                    )
                ]
                ch = model_channels * mult
                if ds in attention_resolutions:
                    if num_head_channels == -1:
                        dim_head = ch // num_heads
                    else:
                        num_heads = ch // num_head_channels
                        dim_head = num_head_channels

                    layers.append(
                        get_attention_layer(
                            ch,
                            num_heads,
                            dim_head,
                            depth=transformer_depth[level],
                            context_dim=context_dim,
                            use_checkpoint=use_checkpoint,
                            disabled_sa=False,
                        )
                    )
                if level and i == num_res_blocks:
                    out_ch = ch
                    ds //= 2
                    layers.append(
                        get_resblock(
                            merge_factor=merge_factor,
                            merge_strategy=merge_strategy,
                            video_kernel_size=video_kernel_size,
                            ch=ch,
                            time_embed_dim=time_embed_dim,
                            dropout=dropout,
                            out_ch=out_ch,
                            dims=dims,
                            use_checkpoint=use_checkpoint,
                            use_scale_shift_norm=use_scale_shift_norm,
                            up=True,
                        )
                        if resblock_updown
                        else Upsample(
                            ch,
                            conv_resample,
                            dims=dims,
                            out_channels=out_ch,
                            third_up=time_downup,
                        )
                    )

                self.output_blocks.append(TimestepEmbedSequential(*layers))
                self._feature_size += ch

        self.out = nn.Sequential(
            normalization(ch),
            nn.SiLU(),
            zero_module(conv_nd(dims, model_channels, out_channels, 3, padding=1)),
        )

    def forward(
        self,
        x: th.Tensor,
        timesteps: th.Tensor,
        context: Optional[th.Tensor] = None,
        y: Optional[th.Tensor] = None,
        time_context: Optional[th.Tensor] = None,
        num_video_frames: Optional[int] = None,
        image_only_indicator: Optional[th.Tensor] = None,
    ):
        assert (
            (y is not None) == (self.num_classes is not None)
        ), "must specify y if and only if the model is class-conditional -> no, relax this TODO"
        hs = []
        t_emb = timestep_embedding(timesteps, self.model_channels, repeat_only=False)
        emb = self.time_embed(t_emb)

        if self.num_classes is not None:
            assert y.shape[0] == x.shape[0]
            emb = emb + self.label_emb(y)

        h = x
        for module in self.input_blocks:
            h = module(
                h,
                emb,
                context=context,
                image_only_indicator=image_only_indicator,
                time_context=time_context,
                num_video_frames=num_video_frames,
            )
            hs.append(h)
        h = self.middle_block(
            h,
            emb,
            context=context,
            image_only_indicator=image_only_indicator,
            time_context=time_context,
            num_video_frames=num_video_frames,
        )
        for module in self.output_blocks:
            h = th.cat([h, hs.pop()], dim=1)
            h = module(
                h,
                emb,
                context=context,
                image_only_indicator=image_only_indicator,
                time_context=time_context,
                num_video_frames=num_video_frames,
            )
        h = h.type(x.dtype)
        return self.out(h)



import imaginairy.vendored.refiners.fluxion.layers as fl
from imaginairy.vendored.refiners.fluxion.layers.attentions import ScaledDotProductAttention
from imaginairy.vendored.refiners.foundationals.latent_diffusion.self_attention_guidance import (
    SAGAdapter,
    SelfAttentionMap,
    SelfAttentionShape,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_xl.unet import MiddleBlock, ResidualBlock, SDXLUNet


class SDXLSAGAdapter(SAGAdapter[SDXLUNet]):
    def __init__(self, target: SDXLUNet, scale: float = 1.0, kernel_size: int = 9, sigma: float = 1.0) -> None:
        super().__init__(
            target=target,
            scale=scale,
            kernel_size=kernel_size,
            sigma=sigma,
        )

    def inject(self: "SDXLSAGAdapter", parent: fl.Chain | None = None) -> "SDXLSAGAdapter":
        middle_block = self.target.ensure_find(MiddleBlock)
        middle_block.insert_after_type(ResidualBlock, SelfAttentionShape(context_key="middle_block_attn_shape"))

        # An alternative would be to replace the ScaledDotProductAttention with a version which records the attention
        # scores to avoid computing these scores twice
        self_attn = middle_block.ensure_find(fl.SelfAttention)
        self_attn.insert_before_type(
            ScaledDotProductAttention,
            SelfAttentionMap(num_heads=self_attn.num_heads, context_key="middle_block_attn_map"),
        )

        return super().inject(parent)

    def eject(self) -> None:
        middle_block = self.target.ensure_find(MiddleBlock)
        middle_block.remove(middle_block.ensure_find(SelfAttentionShape))

        self_attn = middle_block.ensure_find(fl.SelfAttention)
        self_attn.remove(self_attn.ensure_find(SelfAttentionMap))

        super().eject()



import torch
from torch import Size, Tensor, device as Device, dtype as DType, randn
from torch.nn import Parameter as TorchParameter

from imaginairy.vendored.refiners.fluxion.layers.module import Module, WeightedModule


class Identity(Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        return x


class View(Module):
    def __init__(self, *shape: int) -> None:
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        return x.view(*self.shape)


class GetArg(Module):
    def __init__(self, index: int) -> None:
        super().__init__()
        self.index = index

    def forward(self, *args: Tensor) -> Tensor:
        return args[self.index]


class Flatten(Module):
    def __init__(self, start_dim: int = 0, end_dim: int = -1) -> None:
        super().__init__()
        self.start_dim = start_dim
        self.end_dim = end_dim

    def forward(self, x: Tensor) -> Tensor:
        return x.flatten(self.start_dim, self.end_dim)


class Unflatten(Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, x: Tensor, sizes: Size) -> Tensor:
        return x.unflatten(self.dim, sizes)  # type: ignore


class Reshape(Module):
    """
    Reshape the input tensor to the given shape. The shape must be compatible with the input tensor shape. The batch
    dimension is preserved.
    """

    def __init__(self, *shape: int) -> None:
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        return x.reshape(x.shape[0], *self.shape)


class Transpose(Module):
    def __init__(self, dim0: int, dim1: int) -> None:
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1

    def forward(self, x: Tensor) -> Tensor:
        return x.transpose(self.dim0, self.dim1)


class Permute(Module):
    def __init__(self, *dims: int) -> None:
        super().__init__()
        self.dims = dims

    def forward(self, x: Tensor) -> Tensor:
        return x.permute(*self.dims)


class Slicing(Module):
    def __init__(self, dim: int = 0, start: int = 0, end: int | None = None, step: int = 1) -> None:
        super().__init__()
        self.dim = dim
        self.start = start
        self.end = end
        self.step = step

    def forward(self, x: Tensor) -> Tensor:
        dim_size = x.shape[self.dim]
        start = self.start if self.start >= 0 else dim_size + self.start
        end = self.end or dim_size
        end = end if end >= 0 else dim_size + end
        start = max(min(start, dim_size), 0)
        end = max(min(end, dim_size), 0)
        if start >= end:
            return self.get_empty_slice(x)
        indices = torch.arange(start=start, end=end, step=self.step, device=x.device)
        return x.index_select(self.dim, indices)

    def get_empty_slice(self, x: Tensor) -> Tensor:
        """
        Return an empty slice of the same shape as the input tensor to mimic PyTorch's slicing behavior.
        """
        shape = list(x.shape)
        shape[self.dim] = 0
        return torch.empty(*shape, device=x.device)


class Squeeze(Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return x.squeeze(self.dim)


class Unsqueeze(Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return x.unsqueeze(self.dim)


class Unbind(Module):
    def __init__(self, dim: int = 0) -> None:
        self.dim = dim
        super().__init__()

    def forward(self, x: Tensor) -> tuple[Tensor, ...]:
        return x.unbind(dim=self.dim)  # type: ignore


class Chunk(Module):
    def __init__(self, chunks: int, dim: int = 0) -> None:
        self.chunks = chunks
        self.dim = dim
        super().__init__()

    def forward(self, x: Tensor) -> tuple[Tensor, ...]:
        return x.chunk(chunks=self.chunks, dim=self.dim)  # type: ignore


class Sin(Module):
    def forward(self, x: Tensor) -> Tensor:
        return torch.sin(input=x)


class Cos(Module):
    def forward(self, x: Tensor) -> Tensor:
        return torch.cos(input=x)


class Multiply(Module):
    def __init__(self, scale: float = 1.0, bias: float = 0.0) -> None:
        super().__init__()
        self.scale = scale
        self.bias = bias

    def forward(self, x: Tensor) -> Tensor:
        return self.scale * x + self.bias


class Parameter(WeightedModule):
    """
    A layer that wraps a tensor as a parameter. This is useful to create a parameter that is not a weight or a bias.
    """

    def __init__(self, *dims: int, device: Device | str | None = None, dtype: DType | None = None) -> None:
        super().__init__()
        self.dims = dims
        self.weight = TorchParameter(randn(*dims, device=device, dtype=dtype))

    def forward(self, x: Tensor) -> Tensor:
        return self.weight.expand(x.shape[0], *self.dims)


class Buffer(WeightedModule):
    """
    A layer that wraps a tensor as a buffer. This is useful to create a buffer that is not a weight or a bias.

    Buffers are not trainable.
    """

    def __init__(self, *dims: int, device: Device | str | None = None, dtype: DType | None = None) -> None:
        super().__init__()
        self.dims = dims
        self.register_buffer("buffer", randn(*dims, device=device, dtype=dtype))

    @property
    def device(self) -> Device:
        return self.buffer.device

    @property
    def dtype(self) -> DType:
        return self.buffer.dtype

    def forward(self, _: Tensor) -> Tensor:
        return self.buffer



import itertools
import random

from imaginairy.utils.roi_utils import (
    RoiNotInBoundsError,
    resize_roi_coordinates,
    square_roi_coordinate,
)


def test_square_roi_coordinate():
    img_sizes = (10, 100, 200, 511, 513, 1024)
    # iterate through all permutations of image sizes using itertools.product
    for img_width, img_height in itertools.product(img_sizes, img_sizes):
        # randomly generate a region of interest
        for _ in range(100):
            x1 = random.randint(0, img_width)
            y1 = random.randint(0, img_height)
            x2 = random.randint(x1, img_width)
            y2 = random.randint(y1, img_height)
            roi = x1, y1, x2, y2
            try:
                x1, y1, x2, y2 = square_roi_coordinate(roi, img_width, img_height)
            except RoiNotInBoundsError:
                continue
            assert (
                x2 - x1 == y2 - y1
            ), f"ROI is not square: img_width: {img_width}, img_height: {img_height}, roi: {roi}"


# resize_roi_coordinates


def test_square_resize_roi_coordinates():
    img_sizes = (10, 100, 200, 403, 511, 513, 604, 1024)
    # iterate through all permutations of image sizes using itertools.product
    img_sizes = list(itertools.product(img_sizes, img_sizes))

    for img_width, img_height in img_sizes:
        # randomly generate a region of interest
        rois = []
        for _ in range(100):
            x1 = random.randint(0 + 1, img_width - 1)
            y1 = random.randint(0 + 1, img_height - 1)
            x2 = random.randint(x1 + 1, img_width)
            y2 = random.randint(y1 + 1, img_height)
            roi = x1, y1, x2, y2
            rois.append(roi)
        rois.append((392, 85, 695, 389))
        for roi in rois:
            try:
                squared_roi = square_roi_coordinate(roi, img_width, img_height)
            except RoiNotInBoundsError:
                continue
            for n in range(10):
                factor = 1.25 + 0.3 * n
                x1, y1, x2, y2 = resize_roi_coordinates(
                    squared_roi, factor, img_width, img_height
                )
                assert (
                    x2 - x1 == y2 - y1
                ), f"ROI is not square: img_width: {img_width}, img_height: {img_height}, roi: {roi}"

                x1, y1, x2, y2 = resize_roi_coordinates(
                    squared_roi, factor, img_width, img_height, expand_up=False
                )
                assert (
                    x2 - x1 == y2 - y1
                ), f"ROI is not square: img_width: {img_width}, img_height: {img_height}, roi: {roi}"



from typing import Iterable, cast

from torch import Tensor, device as Device, dtype as DType

from imaginairy.vendored.refiners.fluxion.adapters.adapter import Adapter
from imaginairy.vendored.refiners.fluxion.context import Contexts
from imaginairy.vendored.refiners.fluxion.layers import Chain, Conv2d, Lambda, Passthrough, Residual, SiLU, Slicing, UseContext
from imaginairy.vendored.refiners.foundationals.latent_diffusion.range_adapter import RangeAdapter2d
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_1.unet import (
    DownBlocks,
    MiddleBlock,
    ResidualBlock,
    SD1UNet,
    TimestepEncoder,
)


class ConditionEncoder(Chain):
    """Encode an image to be used as a condition for Controlnet.

    Input is a `batch 3 width height` tensor, output is a `batch 320 width//8 height//8` tensor.
    """

    def __init__(self, device: Device | str | None = None, dtype: DType | None = None) -> None:
        self.out_channels = (16, 32, 96, 256)
        super().__init__(
            Chain(
                Conv2d(
                    in_channels=3,
                    out_channels=self.out_channels[0],
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    device=device,
                    dtype=dtype,
                ),
                SiLU(),
            ),
            *(
                Chain(
                    Conv2d(
                        in_channels=self.out_channels[i],
                        out_channels=self.out_channels[i],
                        kernel_size=3,
                        padding=1,
                        device=device,
                        dtype=dtype,
                    ),
                    SiLU(),
                    Conv2d(
                        in_channels=self.out_channels[i],
                        out_channels=self.out_channels[i + 1],
                        kernel_size=3,
                        stride=2,
                        padding=1,
                        device=device,
                        dtype=dtype,
                    ),
                    SiLU(),
                )
                for i in range(len(self.out_channels) - 1)
            ),
            Conv2d(
                in_channels=self.out_channels[-1],
                out_channels=320,
                kernel_size=3,
                padding=1,
                device=device,
                dtype=dtype,
            ),
        )


class Controlnet(Passthrough):
    def __init__(
        self, name: str, scale: float = 1.0, device: Device | str | None = None, dtype: DType | None = None
    ) -> None:
        """Controlnet is a Half-UNet that collects residuals from the UNet and uses them to condition the UNet.

        Input is a `batch 3 width height` tensor, output is a `batch 1280 width//8 height//8` tensor with residuals
        stored in the context.

        It has to use the same context as the UNet: `unet` and `sampling`.
        """
        self.name = name
        self.scale = scale
        super().__init__(
            TimestepEncoder(context_key=f"timestep_embedding_{name}", device=device, dtype=dtype),
            Slicing(dim=1, end=4),  # support inpainting
            DownBlocks(in_channels=4, device=device, dtype=dtype),
            MiddleBlock(device=device, dtype=dtype),
        )

        # We run the condition encoder at each step. Caching the result
        # is not worth it as subsequent runs take virtually no time (FG-374).
        self.DownBlocks[0].append(
            Residual(
                UseContext("controlnet", f"condition_{name}"),
                ConditionEncoder(device=device, dtype=dtype),
            ),
        )
        for residual_block in self.layers(ResidualBlock):
            chain = residual_block.Chain
            RangeAdapter2d(
                target=chain.Conv2d_1,
                channels=residual_block.out_channels,
                embedding_dim=1280,
                context_key=f"timestep_embedding_{name}",
                device=device,
                dtype=dtype,
            ).inject(chain)
        for n, block in enumerate(cast(Iterable[Chain], self.DownBlocks)):
            assert hasattr(block[0], "out_channels"), (
                "The first block of every subchain in DownBlocks is expected to respond to `out_channels`,"
                f" {block[0]} does not."
            )
            out_channels: int = block[0].out_channels
            block.append(
                Passthrough(
                    Conv2d(
                        in_channels=out_channels, out_channels=out_channels, kernel_size=1, device=device, dtype=dtype
                    ),
                    Lambda(self._store_nth_residual(n)),
                )
            )
        self.MiddleBlock.append(
            Passthrough(
                Conv2d(in_channels=1280, out_channels=1280, kernel_size=1, device=device, dtype=dtype),
                Lambda(self._store_nth_residual(12)),
            )
        )

    def _store_nth_residual(self, n: int):
        def _store_residual(x: Tensor):
            residuals = self.use_context("unet")["residuals"]
            residuals[n] = residuals[n] + x * self.scale
            return x

        return _store_residual


class SD1ControlnetAdapter(Chain, Adapter[SD1UNet]):
    def __init__(
        self, target: SD1UNet, name: str, scale: float = 1.0, weights: dict[str, Tensor] | None = None
    ) -> None:
        self.name = name

        controlnet = Controlnet(name=name, scale=scale, device=target.device, dtype=target.dtype)
        if weights is not None:
            controlnet.load_state_dict(weights)
        self._controlnet: list[Controlnet] = [controlnet]  # not registered by PyTorch

        with self.setup_adapter(target):
            super().__init__(target)

    def inject(self: "SD1ControlnetAdapter", parent: Chain | None = None) -> "SD1ControlnetAdapter":
        controlnet = self._controlnet[0]
        target_controlnets = [x for x in self.target if isinstance(x, Controlnet)]
        assert controlnet not in target_controlnets, f"{controlnet} is already injected"
        for cn in target_controlnets:
            assert cn.name != self.name, f"Controlnet named {self.name} is already injected"
        self.target.insert(0, controlnet)
        return super().inject(parent)

    def eject(self) -> None:
        self.target.remove(self._controlnet[0])
        super().eject()

    def init_context(self) -> Contexts:
        return {"controlnet": {f"condition_{self.name}": None}}

    def set_scale(self, scale: float) -> None:
        self._controlnet[0].scale = scale

    def set_controlnet_condition(self, condition: Tensor) -> None:
        self.set_context("controlnet", {f"condition_{self.name}": condition})

    def structural_copy(self: "SD1ControlnetAdapter") -> "SD1ControlnetAdapter":
        raise RuntimeError("Controlnet cannot be copied, eject it first.")



"""Functions for collapsing state_dict key patterns"""


def find_state_dict_key_patterns(patterns):
    """Given a list of state_dict keys, collapse similar keys into patterns.

    For example, if the keys are:

    foo.bar.0.baz
    foo.bar.1.baz

    Then the pattern will be:

    foo.bar.(0|1).baz

    """
    prev_pattern_count = len(patterns) + 1

    # keep running the pattern collapse function until the list of patterns doesn't get any smaller
    while prev_pattern_count > len(patterns):
        prev_pattern_count = len(patterns)
        prev_pattern_count_sub = len(patterns) + 1
        while prev_pattern_count_sub > len(patterns):
            prev_pattern_count_sub = len(patterns)
            patterns = _collapse_patterns(patterns)
        prev_pattern_count_sub = len(patterns) + 1
        while prev_pattern_count_sub > len(patterns):
            prev_pattern_count_sub = len(patterns)
            patterns = _collapse_patterns(patterns, reverse_sort=True)

    return patterns


def prefix_only(k):
    return k.rsplit(".", 1)[0]


def nested_dict_from_keys(keys):
    output = {}
    for key in keys:
        parts = key.split(".")
        # Start from the root of the output and iteratively go deeper
        current_level = output
        for part in parts:
            # If the key doesn't exist at the current level, create a new dict
            if part not in current_level:
                current_level[part] = {}
            # Go one level deeper
            current_level = current_level[part]
    return output


def _collapse_patterns(keys, reverse_sort=False):
    keys = keys.copy()
    keys = [k.split(".") for k in keys]
    if reverse_sort:
        keys.sort(key=lambda k: (len(k), list(reversed(str(k)))))
    else:
        keys.sort(key=lambda k: (len(k), k))
    new_key_patterns = []
    curr_key = None
    for k in keys:
        if curr_key is None:
            curr_key = k
            continue
        single_diff_index = get_single_difference(curr_key, k)
        if single_diff_index is None:
            new_key_patterns.append(curr_key)
            curr_key = k
        else:
            cur_part_val = curr_key[single_diff_index]
            key_part_val = k[single_diff_index]
            if "(" in key_part_val:
                key_vals = key_part_val.strip("()").split("|")
            else:
                key_vals = [key_part_val]
            if "(" in cur_part_val:
                vals = cur_part_val.strip("()").split("|")
            else:
                vals = [cur_part_val]
            vals.extend(key_vals)
            vals.sort()
            try:
                vals = [int(v) for v in vals]
                vals.sort()
                vals = [str(v) for v in vals]
            except ValueError:
                pass
            new_cur_part_val = "(" + "|".join(vals) + ")"
            curr_key[single_diff_index] = new_cur_part_val
    new_key_patterns.append(curr_key)
    new_key_patterns = [".".join(p) for p in new_key_patterns]
    new_key_patterns.sort()
    return new_key_patterns


def get_single_difference(a, b):
    """
    Given two list of strings, if only a single string differs between the two lists, return the index of the differing string.
    """
    if len(a) != len(b):
        return None
    diff_count = 0
    diff_index = None
    for i, (asub, bsub) in enumerate(zip(a, b)):
        if asub != bsub:
            diff_count += 1
            diff_index = i
        if diff_count > 1:
            break

    if diff_count == 1:
        return diff_index
    return None



import numpy as np
import torch
from PIL import Image
from torch import Tensor, device as Device, dtype as DType

from imaginairy.vendored.refiners.fluxion.utils import image_to_tensor, interpolate
from imaginairy.vendored.refiners.foundationals.clip.text_encoder import CLIPTextEncoderL
from imaginairy.vendored.refiners.foundationals.latent_diffusion.auto_encoder import LatentDiffusionAutoencoder
from imaginairy.vendored.refiners.foundationals.latent_diffusion.model import LatentDiffusionModel
from imaginairy.vendored.refiners.foundationals.latent_diffusion.schedulers.dpm_solver import DPMSolver
from imaginairy.vendored.refiners.foundationals.latent_diffusion.schedulers.scheduler import Scheduler
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_1.self_attention_guidance import SD1SAGAdapter
from imaginairy.vendored.refiners.foundationals.latent_diffusion.stable_diffusion_1.unet import SD1UNet


class SD1Autoencoder(LatentDiffusionAutoencoder):
    encoder_scale: float = 0.18215


class StableDiffusion_1(LatentDiffusionModel):
    unet: SD1UNet
    clip_text_encoder: CLIPTextEncoderL

    def __init__(
        self,
        unet: SD1UNet | None = None,
        lda: SD1Autoencoder | None = None,
        clip_text_encoder: CLIPTextEncoderL | None = None,
        scheduler: Scheduler | None = None,
        device: Device | str = "cpu",
        dtype: DType = torch.float32,
    ) -> None:
        unet = unet or SD1UNet(in_channels=4)
        lda = lda or SD1Autoencoder()
        clip_text_encoder = clip_text_encoder or CLIPTextEncoderL()
        scheduler = scheduler or DPMSolver(num_inference_steps=30)

        super().__init__(
            unet=unet,
            lda=lda,
            clip_text_encoder=clip_text_encoder,
            scheduler=scheduler,
            device=device,
            dtype=dtype,
        )

    def compute_clip_text_embedding(self, text: str, negative_text: str = "") -> Tensor:
        conditional_embedding = self.clip_text_encoder(text)
        if text == negative_text:
            return torch.cat(tensors=(conditional_embedding, conditional_embedding), dim=0)

        negative_embedding = self.clip_text_encoder(negative_text or "")
        return torch.cat(tensors=(negative_embedding, conditional_embedding), dim=0)

    def set_unet_context(self, *, timestep: Tensor, clip_text_embedding: Tensor, **_: Tensor) -> None:
        self.unet.set_timestep(timestep=timestep)
        self.unet.set_clip_text_embedding(clip_text_embedding=clip_text_embedding)

    def set_self_attention_guidance(self, enable: bool, scale: float = 1.0) -> None:
        if enable:
            if sag := self._find_sag_adapter():
                sag.scale = scale
            else:
                SD1SAGAdapter(target=self.unet, scale=scale).inject()
        else:
            if sag := self._find_sag_adapter():
                sag.eject()

    def has_self_attention_guidance(self) -> bool:
        return self._find_sag_adapter() is not None

    def _find_sag_adapter(self) -> SD1SAGAdapter | None:
        for p in self.unet.get_parents():
            if isinstance(p, SD1SAGAdapter):
                return p
        return None

    def compute_self_attention_guidance(
        self, x: Tensor, noise: Tensor, step: int, *, clip_text_embedding: Tensor, **kwargs: Tensor
    ) -> Tensor:
        sag = self._find_sag_adapter()
        assert sag is not None

        degraded_latents = sag.compute_degraded_latents(
            scheduler=self.scheduler,
            latents=x,
            noise=noise,
            step=step,
            classifier_free_guidance=True,
        )

        timestep = self.scheduler.timesteps[step].unsqueeze(dim=0)
        negative_embedding, _ = clip_text_embedding.chunk(2)
        self.set_unet_context(timestep=timestep, clip_text_embedding=negative_embedding, **kwargs)
        if "ip_adapter" in self.unet.provider.contexts:
            # this implementation is a bit hacky, it should be refactored in the future
            ip_adapter_context = self.unet.use_context("ip_adapter")
            image_embedding_copy = ip_adapter_context["clip_image_embedding"].clone()
            ip_adapter_context["clip_image_embedding"], _ = ip_adapter_context["clip_image_embedding"].chunk(2)
            degraded_noise = self.unet(degraded_latents)
            ip_adapter_context["clip_image_embedding"] = image_embedding_copy
        else:
            degraded_noise = self.unet(degraded_latents)

        return sag.scale * (noise - degraded_noise)


class StableDiffusion_1_Inpainting(StableDiffusion_1):
    def __init__(
        self,
        unet: SD1UNet | None = None,
        lda: SD1Autoencoder | None = None,
        clip_text_encoder: CLIPTextEncoderL | None = None,
        scheduler: Scheduler | None = None,
        device: Device | str = "cpu",
        dtype: DType = torch.float32,
    ) -> None:
        self.mask_latents: Tensor | None = None
        self.target_image_latents: Tensor | None = None
        super().__init__(
            unet=unet, lda=lda, clip_text_encoder=clip_text_encoder, scheduler=scheduler, device=device, dtype=dtype
        )

    def forward(
        self, x: Tensor, step: int, *, clip_text_embedding: Tensor, condition_scale: float = 7.5, **_: Tensor
    ) -> Tensor:
        assert self.mask_latents is not None
        assert self.target_image_latents is not None
        x = torch.cat(tensors=(x, self.mask_latents, self.target_image_latents), dim=1)
        return super().forward(
            x=x,
            step=step,
            clip_text_embedding=clip_text_embedding,
            condition_scale=condition_scale,
        )

    def set_inpainting_conditions(
        self,
        target_image: Image.Image,
        mask: Image.Image,
        latents_size: tuple[int, int] = (64, 64),
    ) -> tuple[Tensor, Tensor]:
        target_image = target_image.convert(mode="RGB")
        mask = mask.convert(mode="L")

        mask_tensor = torch.tensor(data=np.array(object=mask).astype(dtype=np.float32) / 255.0).to(device=self.device)
        mask_tensor = (mask_tensor > 0.5).unsqueeze(dim=0).unsqueeze(dim=0).to(dtype=self.dtype)
        self.mask_latents = interpolate(x=mask_tensor, factor=torch.Size(latents_size))

        init_image_tensor = image_to_tensor(image=target_image, device=self.device, dtype=self.dtype) * 2 - 1
        masked_init_image = init_image_tensor * (1 - mask_tensor)
        self.target_image_latents = self.lda.encode(x=masked_init_image)

        return self.mask_latents, self.target_image_latents

    def compute_self_attention_guidance(
        self, x: Tensor, noise: Tensor, step: int, *, clip_text_embedding: Tensor, **kwargs: Tensor
    ) -> Tensor:
        sag = self._find_sag_adapter()
        assert sag is not None
        assert self.mask_latents is not None
        assert self.target_image_latents is not None

        degraded_latents = sag.compute_degraded_latents(
            scheduler=self.scheduler,
            latents=x,
            noise=noise,
            step=step,
            classifier_free_guidance=True,
        )
        x = torch.cat(
            tensors=(degraded_latents, self.mask_latents, self.target_image_latents),
            dim=1,
        )

        timestep = self.scheduler.timesteps[step].unsqueeze(dim=0)
        negative_embedding, _ = clip_text_embedding.chunk(2)
        self.set_unet_context(timestep=timestep, clip_text_embedding=negative_embedding, **kwargs)

        if "ip_adapter" in self.unet.provider.contexts:
            # this implementation is a bit hacky, it should be refactored in the future
            ip_adapter_context = self.unet.use_context("ip_adapter")
            image_embedding_copy = ip_adapter_context["clip_image_embedding"].clone()
            ip_adapter_context["clip_image_embedding"], _ = ip_adapter_context["clip_image_embedding"].chunk(2)
            degraded_noise = self.unet(x)
            ip_adapter_context["clip_image_embedding"] = image_embedding_copy
        else:
            degraded_noise = self.unet(x)

        return sag.scale * (noise - degraded_noise)



import torch
import torch.nn as nn
import torchvision


class HopeNet(nn.Module):
    # Hopenet with 3 output layers for yaw, pitch and roll
    # Predicts Euler angles by binning and regression with the expected value
    def __init__(self, block, layers, num_bins):
        super(HopeNet, self).__init__()
        if block == 'resnet':
            block = torchvision.models.resnet.Bottleneck
        self.inplanes = 64
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.AvgPool2d(7)
        self.fc_yaw = nn.Linear(512 * block.expansion, num_bins)
        self.fc_pitch = nn.Linear(512 * block.expansion, num_bins)
        self.fc_roll = nn.Linear(512 * block.expansion, num_bins)

        self.idx_tensor = torch.arange(66).float()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
        return nn.Sequential(*layers)

    @staticmethod
    def softmax_temperature(tensor, temperature):
        result = torch.exp(tensor / temperature)
        result = torch.div(result, torch.sum(result, 1).unsqueeze(1).expand_as(result))
        return result

    def bin2degree(self, predict):
        predict = self.softmax_temperature(predict, 1)
        return torch.sum(predict * self.idx_tensor.type_as(predict), 1) * 3 - 99

    def forward(self, x):
        x = self.relu(self.bn1(self.conv1(x)))
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        pre_yaw = self.fc_yaw(x)
        pre_pitch = self.fc_pitch(x)
        pre_roll = self.fc_roll(x)

        yaw = self.bin2degree(pre_yaw)
        pitch = self.bin2degree(pre_pitch)
        roll = self.bin2degree(pre_roll)
        return yaw, pitch, roll



import pytest

from imaginairy.utils import frange
from imaginairy.utils.prompt_schedules import parse_schedule_str


@pytest.mark.parametrize(
    ("schedule_str", "expected"),
    [
        ("prompt_strength[2:40:1]", ("prompt_strength", list(range(2, 40)))),
        ("prompt_strength[2:40:0.5]", ("prompt_strength", list(frange(2, 40, 0.5)))),
        ("prompt_strength[2,5,10,15]", ("prompt_strength", [2, 5, 10, 15])),
        (
            "prompt_strength[red,blue,10,15]",
            ("prompt_strength", ["red", "blue", 10, 15]),
        ),
    ],
)
def test_parse_schedule_str(schedule_str, expected):
    cleaned_schedules = parse_schedule_str(schedule_str)
    assert cleaned_schedules == expected



"""Functions and class for image generation and encoding"""

import base64
from io import BytesIO

from imaginairy.api import imagine


def generate_image(prompt):
    """ImaginePrompt to generated image."""
    result = next(imagine([prompt]))
    img = result.images["generated"]
    img_io = BytesIO()
    img.save(img_io, "JPEG")
    img_io.seek(0)
    return img_io


def generate_image_b64(prompt):
    """ImaginePrompt to generated base64 encoded image."""
    img_io = generate_image(prompt)
    img_base64 = base64.b64encode(img_io.getvalue())
    return img_base64


class Base64Bytes(bytes):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v, info):
        if isinstance(v, bytes):
            return v
        if isinstance(v, str):
            return base64.b64decode(v)
        raise ValueError("Byte value must be either str or bytes")

    def __str__(self):
        return base64.b64encode(self).decode()



"""Functions for generating synthetic videos"""

import logging
import math
import os
import random
import re
import time
from glob import glob
from pathlib import Path
from typing import Any, Optional

import cv2
import numpy as np
import torch
from einops import rearrange, repeat
from omegaconf import OmegaConf
from PIL import Image
from torchvision.transforms import ToTensor

from imaginairy import config
from imaginairy.enhancers.video_interpolation.rife.interpolate import interpolate_images
from imaginairy.schema import LazyLoadingImage
from imaginairy.utils import (
    default,
    get_device,
    instantiate_from_config,
    platform_appropriate_autocast,
)
from imaginairy.utils.animations import make_bounce_animation
from imaginairy.utils.downloads import get_cached_url_path
from imaginairy.utils.named_resolutions import normalize_image_size
from imaginairy.utils.paths import PKG_ROOT

logger = logging.getLogger(__name__)


def generate_video(
    input_path: str,  # Can either be image file or folder with image files
    output_folder: str | None = None,
    size=(1024, 576),
    num_frames: int = 6,
    num_steps: int = 30,
    model_name: str = "svd-xt",
    fps_id: int = 6,
    output_fps: int = 6,
    motion_bucket_id: int = 127,
    cond_aug: float = 0.02,
    seed: Optional[int] = None,
    decoding_t: int = 1,  # Number of frames decoded at a time! This eats most VRAM. Reduce if necessary.
    device: Optional[str] = None,
    repetitions=1,
    output_format="webp",
):
    """
    Generates a video from a single image or multiple images, conditioned on the provided input_path.

    Args:
        input_path (str): Path to an image file or a directory containing image files.
        output_folder (str | None, optional): Directory where the generated video will be saved.
            Defaults to "outputs/video/" if None.
        num_frames (int, optional): Number of frames in the generated video. Defaults to 6.
        num_steps (int, optional): Number of steps for the generation process. Defaults to 30.
        model_name (str, optional): Name of the model to use for generation. Defaults to "svd_xt".
        fps_id (int, optional): Frame rate identifier used in generation. Defaults to 6.
        output_fps (int, optional): Frame rate of the output video. Defaults to 6.
        motion_bucket_id (int, optional): Identifier for motion bucket. Defaults to 127.
        cond_aug (float, optional): Conditional augmentation value. Defaults to 0.02.
        seed (Optional[int], optional): Random seed for generation. If None, a random seed is chosen.
        decoding_t (int, optional): Number of frames decoded at a time, affecting VRAM usage.
            Reduce if necessary. Defaults to 1.
        device (Optional[str], optional): Device to run the generation on. Defaults to the detected device.
        repetitions (int, optional): Number of times to repeat the video generation process. Defaults to 1.

    Returns:
        None: The function saves the generated video(s) to the specified output folder.
    """
    device = default(device, get_device)
    vid_width, vid_height = normalize_image_size(size)
    if device == "mps":
        msg = "Apple Silicon MPS (M1, M2, etc) is not currently supported for video generation. Switching to cpu generation."
        logger.warning(msg)
        device = "cpu"

    elif not torch.cuda.is_available():
        msg = (
            "CUDA is not available. This will be verrrry slow or not work at all.\n"
            "If you have a GPU, make sure you have CUDA installed and PyTorch is compiled with CUDA support.\n"
            "Unfortunately, we cannot automatically install the proper version.\n\n"
            "You can install the proper version by following these directions:\n"
            "https://pytorch.org/get-started/locally/"
        )
        logger.warning(msg)

    output_fps = default(output_fps, fps_id)

    video_model_config = config.MODEL_WEIGHT_CONFIG_LOOKUP.get(model_name, None)
    if video_model_config is None:
        msg = f"Version {model_name} does not exist."
        raise ValueError(msg)

    num_frames = default(num_frames, video_model_config.defaults.get("frames", 12))
    num_steps = default(num_steps, video_model_config.defaults.get("steps", 30))
    output_folder_str = default(output_folder, "outputs/video/")
    del output_folder
    video_config_path = f"{PKG_ROOT}/{video_model_config.architecture.config_path}"

    model, safety_filter = load_model(
        config=video_config_path,
        device="cpu",
        num_frames=num_frames,
        num_steps=num_steps,
        weights_url=video_model_config.weights_location,
    )

    if input_path.startswith("http"):
        all_img_paths = [input_path]
    else:
        path = Path(input_path)
        if path.is_file():
            if any(input_path.endswith(x) for x in ["jpg", "jpeg", "png"]):
                all_img_paths = [input_path]
            else:
                raise ValueError("Path is not valid image file.")
        elif path.is_dir():
            all_img_paths = sorted(
                [
                    str(f)
                    for f in path.iterdir()
                    if f.is_file() and f.suffix.lower() in [".jpg", ".jpeg", ".png"]
                ]
            )
            if len(all_img_paths) == 0:
                raise ValueError("Folder does not contain any images.")
        else:
            msg = f"Could not find file or folder at {input_path}"
            raise FileNotFoundError(msg)

    expected_size = (vid_width, vid_height)
    for _ in range(repetitions):
        for input_path in all_img_paths:
            start_time = time.perf_counter()
            _seed = default(seed, random.randint(0, 1000000))
            torch.manual_seed(_seed)
            logger.info(
                f"Generating a {num_frames} frame video from {input_path}. Device:{device} seed:{_seed}"
            )
            if input_path.startswith("http"):
                image = LazyLoadingImage(url=input_path).as_pillow()
            else:
                image = LazyLoadingImage(filepath=input_path).as_pillow()
            crop_coords = None
            if image.mode == "RGBA":
                image = image.convert("RGB")
            if image.size != expected_size:
                logger.info(
                    f"Resizing image from {image.size} to {expected_size}. (w, h)"
                )
                image = pillow_fit_image_within(
                    image, max_height=expected_size[1], max_width=expected_size[0]
                )
                logger.debug(f"Image is now of size: {image.size}")
                background = Image.new("RGB", expected_size, "white")
                # Calculate the position to center the original image
                x = (background.width - image.width) // 2
                y = (background.height - image.height) // 2
                background.paste(image, (x, y))
                # crop_coords = (x, y, x + image.width, y + image.height)

                # image = background
            w, h = image.size
            snap_to = 64
            if h % snap_to != 0 or w % snap_to != 0:
                width = w - w % snap_to
                height = h - h % snap_to
                image = image.resize((width, height))
                logger.warning(
                    f"Your image is of size {h}x{w} which is not divisible by 64. We are resizing to {height}x{width}!"
                )

            image = ToTensor()(image)
            image = image * 2.0 - 1.0

            image = image.unsqueeze(0).to(device)
            H, W = image.shape[2:]
            assert image.shape[1] == 3
            F = 8
            C = 4
            shape = (num_frames, C, H // F, W // F)
            if expected_size != (W, H):
                logger.warning(
                    f"The {W, H} image you provided is not {expected_size}.  This leads to suboptimal performance as model was only trained on 576x1024. Consider increasing `cond_aug`."
                )
            if motion_bucket_id > 255:
                logger.warning(
                    "High motion bucket! This may lead to suboptimal performance."
                )

            if fps_id < 5:
                logger.warning(
                    "Small fps value! This may lead to suboptimal performance."
                )

            if fps_id > 30:
                logger.warning(
                    "Large fps value! This may lead to suboptimal performance."
                )

            value_dict: dict[str, Any] = {}
            value_dict["motion_bucket_id"] = motion_bucket_id
            value_dict["fps_id"] = fps_id
            value_dict["cond_aug"] = cond_aug
            value_dict["cond_frames_without_noise"] = image
            value_dict["cond_frames"] = image + cond_aug * torch.randn_like(image)

            with torch.no_grad(), platform_appropriate_autocast():
                reload_model(model.conditioner, device=device)
                if device == "cpu":
                    model.conditioner.to(torch.float32)
                for k in value_dict:
                    if isinstance(value_dict[k], torch.Tensor):
                        value_dict[k] = value_dict[k].to(
                            next(model.conditioner.parameters()).dtype
                        )
                batch, batch_uc = get_batch(
                    get_unique_embedder_keys_from_conditioner(model.conditioner),
                    value_dict,
                    [1, num_frames],
                    T=num_frames,
                    device=device,
                )
                c, uc = model.conditioner.get_unconditional_conditioning(
                    batch,
                    batch_uc=batch_uc,
                    force_uc_zero_embeddings=[
                        "cond_frames",
                        "cond_frames_without_noise",
                    ],
                )
                unload_model(model.conditioner)

                for k in ["crossattn", "concat"]:
                    uc[k] = repeat(uc[k], "b ... -> b t ...", t=num_frames)
                    uc[k] = rearrange(uc[k], "b t ... -> (b t) ...", t=num_frames)
                    c[k] = repeat(c[k], "b ... -> b t ...", t=num_frames)
                    c[k] = rearrange(c[k], "b t ... -> (b t) ...", t=num_frames)

                randn = torch.randn(shape, device=device, dtype=torch.float16)

                additional_model_inputs = {}
                additional_model_inputs["image_only_indicator"] = torch.zeros(
                    2, num_frames
                ).to(device)
                additional_model_inputs["num_video_frames"] = batch["num_video_frames"]

                def denoiser(_input, sigma, c):
                    _input = _input.half().to(device)
                    return model.denoiser(
                        model.model, _input, sigma, c, **additional_model_inputs
                    )

                reload_model(model.denoiser, device=device)
                reload_model(model.model, device=device)
                samples_z = model.sampler(denoiser, randn, cond=c, uc=uc)
                unload_model(model.model)
                unload_model(model.denoiser)

                reload_model(model.first_stage_model, device=device)
                model.en_and_decode_n_samples_a_time = decoding_t
                samples_x = model.decode_first_stage(samples_z)
                samples = torch.clamp((samples_x + 1.0) / 2.0, min=0.0, max=1.0)
                unload_model(model.first_stage_model)

                if crop_coords:
                    left, upper, right, lower = crop_coords
                    samples = samples[:, :, upper:lower, left:right]

                os.makedirs(output_folder_str, exist_ok=True)
                base_count = len(glob(os.path.join(output_folder_str, "*.*"))) + 1
                source_slug = make_safe_filename(input_path)
                video_filename = f"{base_count:06d}_{model_name}_{_seed}_{fps_id}fps_{source_slug}.{output_format}"
                video_path = os.path.join(output_folder_str, video_filename)

                samples = safety_filter(samples)
                # save_video(samples, video_path, output_fps)
                save_video_bounce(samples, video_path, output_fps)

            duration = time.perf_counter() - start_time
            logger.info(
                f"Video of {num_frames} frames generated in {duration:.2f} seconds and saved to {video_path}\n"
            )


def save_video(samples: torch.Tensor, video_filename: str, output_fps: int):
    """
    Saves a video from given tensor samples.

    Args:
    samples (torch.Tensor): Tensor containing video frame data.
    video_filename (str): The full path and filename where the video will be saved.
    output_fps (int): Frames per second for the output video.
    safety_filter (Callable[[torch.Tensor], torch.Tensor]): A function to apply a safety filter to the samples.

    Returns:
    str: The path to the saved video.
    """
    vid = (torch.permute(samples, (0, 2, 3, 1)) * 255).cpu().numpy().astype(np.uint8)
    writer = cv2.VideoWriter(
        video_filename,
        cv2.VideoWriter_fourcc(*"MP4V"),  # type: ignore
        output_fps,
        (samples.shape[-1], samples.shape[-2]),
    )
    for frame in vid:
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        writer.write(frame)
    writer.release()
    video_path_h264 = video_filename[:-4] + "_h264.mp4"
    os.system(f"ffmpeg -i {video_filename} -c:v libx264 {video_path_h264}")


def save_video_bounce(
    samples: torch.Tensor, video_filename: str, output_fps: int, interpolate_fps=60
):
    frames_np = (
        (torch.permute(samples, (0, 2, 3, 1)) * 255).cpu().numpy().astype(np.uint8)
    )
    transition_duration = len(frames_np) / float(output_fps)
    frames_pil = [Image.fromarray(frame) for frame in frames_np]
    if interpolate_fps:
        # bring it up to at least 60 fps
        fps_multiplier = int(math.ceil(interpolate_fps / output_fps))
        frames_pil = interpolate_images(frames_pil, fps_multiplier=fps_multiplier)

    transition_duration_ms = transition_duration * 1000
    logger.info(
        f"Interpolated from {len(frames_np)} to {len(frames_pil)} frames ({fps_multiplier} multiplier)"
    )
    logger.info(
        f"Making bounce animation with transition duration {transition_duration_ms:.1f}ms"
    )
    make_bounce_animation(
        imgs=frames_pil,
        outpath=video_filename,
        transition_duration_ms=transition_duration_ms,
        end_pause_duration_ms=100,
        max_fps=60,
    )


def get_unique_embedder_keys_from_conditioner(conditioner):
    return list({x.input_key for x in conditioner.embedders})


def get_batch(keys, value_dict, N, T, device):
    batch = {}
    batch_uc = {}

    for key in keys:
        if key == "fps_id":
            batch[key] = (
                torch.tensor([value_dict["fps_id"]])
                .to(device)
                .repeat(int(math.prod(N)))
            )
        elif key == "motion_bucket_id":
            batch[key] = (
                torch.tensor([value_dict["motion_bucket_id"]])
                .to(device)
                .repeat(int(math.prod(N)))
            )
        elif key == "cond_aug":
            batch[key] = repeat(
                torch.tensor([value_dict["cond_aug"]]).to(device),
                "1 -> b",
                b=math.prod(N),
            )
        elif key == "cond_frames":
            batch[key] = repeat(value_dict["cond_frames"], "1 ... -> b ...", b=N[0])
        elif key == "cond_frames_without_noise":
            batch[key] = repeat(
                value_dict["cond_frames_without_noise"], "1 ... -> b ...", b=N[0]
            )
        else:
            batch[key] = value_dict[key]

    if T is not None:
        batch["num_video_frames"] = T

    for key in batch:
        if key not in batch_uc and isinstance(batch[key], torch.Tensor):
            batch_uc[key] = torch.clone(batch[key])
    return batch, batch_uc


def load_model(
    config: str, device: str, num_frames: int, num_steps: int, weights_url: str
):
    oconfig = OmegaConf.load(config)
    ckpt_path = get_cached_url_path(weights_url)
    oconfig["model"]["params"]["ckpt_path"] = ckpt_path  # type: ignore
    if device == "cuda":
        oconfig.model.params.conditioner_config.params.emb_models[
            0
        ].params.open_clip_embedding_config.params.init_device = device

    oconfig.model.params.sampler_config.params.num_steps = num_steps
    oconfig.model.params.sampler_config.params.guider_config.params.num_frames = (
        num_frames
    )

    model = instantiate_from_config(oconfig.model).to(device).half().eval()

    # safety_filter = DeepFloydDataFiltering(verbose=False, device=device)
    def safety_filter(x):
        return x

    # use less memory
    model.model.half()
    return model, safety_filter


lowvram_mode = True


def unload_model(model):
    global lowvram_mode
    if lowvram_mode:
        model.cpu()
        if get_device() == "cuda":
            torch.cuda.empty_cache()


def reload_model(model, device=None):
    device = default(device, get_device)
    model.to(device)


def pillow_fit_image_within(
    image: Image.Image, max_height=512, max_width=512, convert="RGB", snap_size=8
):
    image = image.convert(convert)
    w, h = image.size
    resize_ratio = 1

    if w > max_width or h > max_height:
        resize_ratio = min(max_width / w, max_height / h)
    elif w < max_width and h < max_height:
        # it's smaller than our target image, enlarge
        resize_ratio = min(max_width / w, max_height / h)

    if resize_ratio != 1:
        w, h = int(w * resize_ratio), int(h * resize_ratio)
    # resize to integer multiple of snap_size
    w -= w % snap_size
    h -= h % snap_size

    if (w, h) != image.size:
        image = image.resize((w, h), resample=Image.Resampling.LANCZOS)
    return image


def make_safe_filename(input_string):
    stripped_url = re.sub(r"^https?://[^/]+/", "", input_string)

    # Remove directory path if present
    base_name = os.path.basename(stripped_url)

    # Remove file extension
    name_without_extension = os.path.splitext(base_name)[0]

    # Keep only alphanumeric characters and dashes
    safe_name = re.sub(r"[^a-zA-Z0-9\-]", "", name_without_extension)

    return safe_name



"""Classes and functions for managing AI models"""

import logging
import os
import sys
from functools import lru_cache

import torch
from omegaconf import OmegaConf
from safetensors.torch import load_file

from imaginairy import config as iconfig
from imaginairy.config import IMAGE_WEIGHTS_SHORT_NAMES, ModelArchitecture
from imaginairy.modules import attention
from imaginairy.modules.refiners_sd import (
    SDXLAutoencoderSliced,
    StableDiffusion_XL,
    StableDiffusion_XL_Inpainting,
)
from imaginairy.utils import clear_gpu_cache, get_device, instantiate_from_config
from imaginairy.utils.downloads import (
    HuggingFaceAuthorizationError,
    download_huggingface_weights,
    get_cached_url_path,
    is_diffusers_repo_url,
    normalize_diffusers_repo_url,
)
from imaginairy.utils.model_cache import memory_managed_model
from imaginairy.utils.named_resolutions import normalize_image_size
from imaginairy.utils.paths import PKG_ROOT
from imaginairy.vendored.refiners.foundationals.clip.text_encoder import (
    CLIPTextEncoderL,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion import (
    DoubleTextEncoder,
    SD1UNet,
    SDXLUNet,
)
from imaginairy.vendored.refiners.foundationals.latent_diffusion.model import (
    LatentDiffusionModel,
)
from imaginairy.weight_management import translators
from imaginairy.weight_management.translators import (
    DoubleTextEncoderTranslator,
    diffusers_autoencoder_kl_to_refiners_translator,
    diffusers_unet_sdxl_to_refiners_translator,
    load_weight_map,
)

logger = logging.getLogger(__name__)

MOST_RECENTLY_LOADED_MODEL = None


def load_state_dict(weights_location, half_mode=False, device=None):
    if device is None:
        device = get_device()

    if weights_location.startswith("http"):
        ckpt_path = get_cached_url_path(weights_location, category="weights")
    else:
        ckpt_path = weights_location
    logger.info(f"Loading model {ckpt_path} onto {get_device()} backend...")
    state_dict = None
    # weights_cache_key = (ckpt_path, half_mode)
    # if weights_cache_key in GLOBAL_WEIGHTS_CACHE:
    #     return GLOBAL_WEIGHTS_CACHE.get(weights_cache_key)

    try:
        state_dict = load_tensors(ckpt_path, map_location="cpu")
    except FileNotFoundError as e:
        if e.errno == 2:
            logger.error(
                f'Error: "{ckpt_path}" not a valid path to model weights.\nPreconfigured models you can use: {IMAGE_WEIGHTS_SHORT_NAMES}.'
            )
            sys.exit(1)
        raise
    except RuntimeError as e:
        err_str = str(e)
        if (
            "PytorchStreamReader failed reading zip archive" in err_str
            and weights_location.startswith("http")
        ):
            logger.warning("Corrupt checkpoint. deleting and re-downloading...")
            os.remove(ckpt_path)
            ckpt_path = get_cached_url_path(weights_location, category="weights")
            state_dict = load_tensors(ckpt_path, map_location="cpu")
        if state_dict is None:
            raise

    state_dict = state_dict.get("state_dict", state_dict)

    if half_mode:
        state_dict = {k: v.half() for k, v in state_dict.items()}

    # change device
    state_dict = {k: v.to(device) for k, v in state_dict.items()}

    # GLOBAL_WEIGHTS_CACHE.set(weights_cache_key, state_dict)

    return state_dict


def load_model_from_config(config, weights_location, half_mode=False):
    model = instantiate_from_config(config.model)
    base_model_dict = load_state_dict(weights_location, half_mode=half_mode)
    model.init_from_state_dict(base_model_dict)
    if half_mode:
        model = model.half()
    model.to(get_device())
    model.eval()
    return model


def get_diffusion_model(
    weights_location=iconfig.DEFAULT_MODEL_WEIGHTS,
    config_path="configs/stable-diffusion-v1.yaml",
    control_weights_locations=None,
    half_mode=None,
    for_inpainting=False,
):
    """
    Load a diffusion model.

    Weights location may also be shortcut name, e.g. "SD-1.5"
    """
    try:
        return _get_diffusion_model(
            weights_location,
            config_path,
            half_mode,
            for_inpainting,
            control_weights_locations=control_weights_locations,
        )
    except HuggingFaceAuthorizationError as e:
        if for_inpainting:
            logger.warning(
                f"Failed to load inpainting model. Attempting to fall-back to standard model.   {e!s}"
            )
            return _get_diffusion_model(
                iconfig.DEFAULT_MODEL_WEIGHTS,
                config_path,
                half_mode,
                for_inpainting=False,
                control_weights_locations=control_weights_locations,
            )
        raise


def _get_diffusion_model(
    weights_location=iconfig.DEFAULT_MODEL_WEIGHTS,
    model_architecture="configs/stable-diffusion-v1.yaml",
    half_mode=None,
    for_inpainting=False,
    control_weights_locations=None,
):
    """
    Load a diffusion model.

    Weights location may also be shortcut name, e.g. "SD-1.5"
    """
    global MOST_RECENTLY_LOADED_MODEL

    model_weights_config = resolve_model_weights_config(
        model_weights=weights_location,
        default_model_architecture=model_architecture,
        for_inpainting=for_inpainting,
    )
    # some models need the attention calculated in float32
    if model_weights_config is not None:
        attention.ATTENTION_PRECISION_OVERRIDE = (
            model_weights_config.forced_attn_precision
        )
    else:
        attention.ATTENTION_PRECISION_OVERRIDE = "default"
    diffusion_model = _load_diffusion_model(
        config_path=model_weights_config.architecture.config_path,
        weights_location=weights_location,
        half_mode=half_mode,
    )
    MOST_RECENTLY_LOADED_MODEL = diffusion_model
    if control_weights_locations:
        controlnets = []
        for control_weights_location in control_weights_locations:
            controlnets.append(load_controlnet(control_weights_location, half_mode))
        diffusion_model.set_control_models(controlnets)

    return diffusion_model


def get_diffusion_model_refiners(
    weights_config: iconfig.ModelWeightsConfig,
    for_inpainting=False,
    dtype=None,
) -> LatentDiffusionModel:
    """Load a diffusion model."""

    sd = _get_diffusion_model_refiners(
        weights_location=weights_config.weights_location,
        architecture_alias=weights_config.architecture.primary_alias,
        for_inpainting=for_inpainting,
        dtype=dtype,
    )
    # ensures a "fresh" copy that doesn't have additional injected parts
    sd = sd.structural_copy()

    sd.set_self_attention_guidance(enable=True)

    return sd


@lru_cache(maxsize=1)
def _get_diffusion_model_refiners(
    weights_location: str,
    architecture_alias: str,
    for_inpainting: bool = False,
    device=None,
    dtype=torch.float16,
) -> LatentDiffusionModel:
    """
    Load a diffusion model.

    Weights location may also be shortcut name, e.g. "SD-1.5"
    """

    global MOST_RECENTLY_LOADED_MODEL
    _get_diffusion_model_refiners.cache_clear()
    clear_gpu_cache()

    architecture = iconfig.MODEL_ARCHITECTURE_LOOKUP[architecture_alias]
    if architecture.primary_alias in ("sd15", "sd15inpaint"):
        sd = load_sd15_pipeline(
            weights_location=weights_location,
            for_inpainting=for_inpainting,
            device=device,
            dtype=dtype,
        )
    elif architecture.primary_alias in ("sdxl", "sdxlinpaint"):
        sd = load_sdxl_pipeline(
            base_url=weights_location, device=device, for_inpainting=for_inpainting
        )
    else:
        msg = f"Invalid architecture {architecture.primary_alias}"
        raise ValueError(msg)

    MOST_RECENTLY_LOADED_MODEL = sd

    msg = (
        "Pipeline loaded "
        f"sd[dtype:{sd.dtype} device:{sd.device}] "
        f"sd.unet[dtype:{sd.unet.dtype} device:{sd.unet.device}] "
        f"sd.lda[dtype:{sd.lda.dtype} device:{sd.lda.device}]"
        f"sd.clip_text_encoder[dtype:{sd.clip_text_encoder.dtype} device:{sd.clip_text_encoder.device}]"
    )
    logger.debug(msg)

    return sd


# new
def load_sd15_pipeline(
    weights_location: str,
    for_inpainting: bool = False,
    device=None,
    dtype=torch.float16,
) -> LatentDiffusionModel:
    """
    Load a diffusion model.

    Weights location may also be shortcut name, e.g. "SD-1.5"
    """
    from imaginairy.modules.refiners_sd import (
        SD1AutoencoderSliced,
        StableDiffusion_1,
        StableDiffusion_1_Inpainting,
    )

    device = device or get_device()
    if is_diffusers_repo_url(weights_location):
        (
            vae_weights,
            unet_weights,
            text_encoder_weights,
        ) = load_sd15_diffusers_weights(weights_location, device="cpu")
    else:
        (
            vae_weights,
            unet_weights,
            text_encoder_weights,
        ) = load_stable_diffusion_compvis_weights(weights_location)
    StableDiffusionCls: type[LatentDiffusionModel]
    if for_inpainting:
        unet = SD1UNet(in_channels=9, device="cpu", dtype=dtype)
        StableDiffusionCls = StableDiffusion_1_Inpainting
    else:
        unet = SD1UNet(in_channels=4, device="cpu", dtype=dtype)
        StableDiffusionCls = StableDiffusion_1
    logger.debug(f"Using class {StableDiffusionCls.__name__}")

    sd = StableDiffusionCls(
        device=device, dtype=dtype, lda=SD1AutoencoderSliced(), unet=unet
    )
    logger.debug("Loading VAE")
    sd.lda.load_state_dict(vae_weights, assign=True)

    logger.debug("Loading text encoder")
    sd.clip_text_encoder.load_state_dict(text_encoder_weights, assign=True)

    logger.debug("Loading UNet")
    sd.unet.load_state_dict(unet_weights, strict=False, assign=True)

    logger.debug(f"'{weights_location}' Loaded")
    sd.to(device=device, dtype=dtype)
    return sd


def _get_sd15_diffusion_model_refiners_new(
    weights_location: str,
    for_inpainting: bool = False,
    device=None,
    dtype=torch.float16,
) -> LatentDiffusionModel:
    """
    Load a diffusion model.

    Weights location may also be shortcut name, e.g. "SD-1.5"
    """
    from imaginairy.modules.refiners_sd import (
        SD1AutoencoderSliced,
        StableDiffusion_1,
        StableDiffusion_1_Inpainting,
    )

    device = device or get_device()
    if is_diffusers_repo_url(weights_location):
        (
            vae_weights,
            unet_weights,
            text_encoder_weights,
        ) = load_sd15_diffusers_weights(weights_location, device="cpu")
    else:
        (
            vae_weights,
            unet_weights,
            text_encoder_weights,
        ) = load_stable_diffusion_compvis_weights(weights_location)
    StableDiffusionCls: type[LatentDiffusionModel]
    if for_inpainting:
        unet = SD1UNet(in_channels=9, device="cpu", dtype=dtype)
        StableDiffusionCls = StableDiffusion_1_Inpainting
    else:
        unet = SD1UNet(in_channels=4, device="cpu", dtype=dtype)
        StableDiffusionCls = StableDiffusion_1

    logger.debug("Loading UNet")
    unet.load_state_dict(unet_weights, strict=False, assign=True)
    del unet_weights
    unet.to(device=device, dtype=dtype)

    logger.debug("Loading VAE")
    lda = SD1AutoencoderSliced(device=device, dtype=dtype)
    lda.load_state_dict(vae_weights, assign=True)
    del vae_weights
    lda.to(device=device, dtype=dtype)

    logger.debug("Loading text encoder")
    clip_text_encoder = CLIPTextEncoderL()
    clip_text_encoder.load_state_dict(text_encoder_weights, assign=True)
    del text_encoder_weights
    clip_text_encoder.to(device=device, dtype=dtype)

    logger.debug(f"Using class {StableDiffusionCls.__name__}")

    sd = StableDiffusionCls(device=None, dtype=dtype, lda=lda, unet=unet)  # type: ignore
    sd.to(device=device, dtype=dtype)

    logger.debug(f"'{weights_location}' Loaded")
    return sd


@memory_managed_model("stable-diffusion", memory_usage_mb=1951)
def _load_diffusion_model(config_path, weights_location, half_mode):
    model_config = OmegaConf.load(f"{PKG_ROOT}/{config_path}")

    # only run half-mode on cuda. run it by default
    half_mode = half_mode is None and get_device() == "cuda"

    model = load_model_from_config(
        config=model_config,
        weights_location=weights_location,
        half_mode=half_mode,
    )
    return model


@memory_managed_model("controlnet")
def load_controlnet(control_weights_location, half_mode):
    controlnet_state_dict = load_state_dict(
        control_weights_location, half_mode=half_mode
    )
    controlnet_state_dict = {
        k.replace("control_model.", ""): v for k, v in controlnet_state_dict.items()
    }
    control_stage_config = OmegaConf.load(f"{PKG_ROOT}/configs/control-net-v15.yaml")[
        "model"
    ]["params"]["control_stage_config"]
    controlnet = instantiate_from_config(control_stage_config)
    controlnet.load_state_dict(controlnet_state_dict, assign=True)
    controlnet.to(get_device())
    return controlnet


def resolve_model_weights_config(
    model_weights: str | iconfig.ModelWeightsConfig,
    default_model_architecture: str | None = None,
    for_inpainting: bool = False,
) -> iconfig.ModelWeightsConfig:
    """Resolve weight and config path if they happen to be shortcuts."""
    if isinstance(model_weights, iconfig.ModelWeightsConfig):
        return model_weights

    if not isinstance(model_weights, str):
        msg = f"Invalid model weights: {model_weights}"
        raise ValueError(msg)  # noqa

    if default_model_architecture is not None and not isinstance(
        default_model_architecture, str
    ):
        msg = f"Invalid model architecture: {default_model_architecture}"
        raise ValueError(msg)

    if for_inpainting:
        model_weights_config = iconfig.MODEL_WEIGHT_CONFIG_LOOKUP.get(
            f"{model_weights.lower()}-inpaint", None
        )
        if model_weights_config:
            return model_weights_config

    model_weights_config = iconfig.MODEL_WEIGHT_CONFIG_LOOKUP.get(
        model_weights.lower(), None
    )
    if model_weights_config:
        return model_weights_config

    if not default_model_architecture:
        msg = "You must specify the model architecture when loading custom weights."
        raise ValueError(msg)

    default_model_architecture = default_model_architecture.lower()
    model_architecture_config = None
    if for_inpainting:
        model_architecture_config = iconfig.MODEL_ARCHITECTURE_LOOKUP.get(
            f"{default_model_architecture}-inpaint", None
        )

    if not model_architecture_config:
        model_architecture_config = iconfig.MODEL_ARCHITECTURE_LOOKUP.get(
            default_model_architecture, None
        )

    if model_architecture_config is None:
        msg = f"Invalid model architecture: {default_model_architecture}"
        raise ValueError(msg)

    model_weights_config = iconfig.ModelWeightsConfig(
        name="Custom Loaded",
        aliases=[],
        architecture=model_architecture_config,
        weights_location=model_weights,
        defaults={},
    )

    return model_weights_config


def get_model_default_image_size(model_architecture: str | ModelArchitecture | None):
    if isinstance(model_architecture, str):
        model_architecture = iconfig.MODEL_ARCHITECTURE_LOOKUP.get(
            model_architecture, None
        )
    default_size = None
    if model_architecture:
        default_size = model_architecture.defaults.get("size")

    if default_size is None:
        default_size = 512
    default_size = normalize_image_size(default_size)
    return default_size


def get_current_diffusion_model():
    return MOST_RECENTLY_LOADED_MODEL


def load_sd15_diffusers_weights(base_url: str, device=None):
    from imaginairy.utils import get_device
    from imaginairy.weight_management.conversion import cast_weights
    from imaginairy.weight_management.utils import (
        COMPONENT_NAMES,
        FORMAT_NAMES,
        MODEL_NAMES,
    )

    base_url = normalize_diffusers_repo_url(base_url)
    if device is None:
        device = get_device()
    vae_weights_path = download_huggingface_weights(base_url=base_url, sub="vae")
    vae_weights = open_weights(vae_weights_path, device=device)
    vae_weights = cast_weights(
        source_weights=vae_weights,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.VAE,
        source_format=FORMAT_NAMES.DIFFUSERS,
        dest_format=FORMAT_NAMES.REFINERS,
    )

    unet_weights_path = download_huggingface_weights(base_url=base_url, sub="unet")
    unet_weights = open_weights(unet_weights_path, device=device)
    unet_weights = cast_weights(
        source_weights=unet_weights,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.UNET,
        source_format=FORMAT_NAMES.DIFFUSERS,
        dest_format=FORMAT_NAMES.REFINERS,
    )

    text_encoder_weights_path = download_huggingface_weights(
        base_url=base_url, sub="text_encoder"
    )
    text_encoder_weights = open_weights(text_encoder_weights_path, device=device)
    text_encoder_weights = cast_weights(
        source_weights=text_encoder_weights,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.TEXT_ENCODER,
        source_format=FORMAT_NAMES.DIFFUSERS,
        dest_format=FORMAT_NAMES.REFINERS,
    )
    first_vae = next(iter(vae_weights.values()))
    first_unet = next(iter(unet_weights.values()))
    first_encoder = next(iter(text_encoder_weights.values()))
    msg = (
        f"vae weights. dtype: {first_vae.dtype} device: {first_vae.device}\n"
        f"unet weights. dtype: {first_unet.dtype} device: {first_unet.device}\n"
        f"text_encoder weights. dtype: {first_encoder.dtype} device: {first_encoder.device}\n"
    )
    logger.debug(msg)
    return vae_weights, unet_weights, text_encoder_weights


def load_sdxl_pipeline_from_diffusers_weights(
    base_url: str, for_inpainting=False, device=None, dtype=torch.float16
):
    from imaginairy.utils import get_device

    device = device or get_device()

    base_url = normalize_diffusers_repo_url(base_url)

    translator = translators.diffusers_autoencoder_kl_to_refiners_translator()
    vae_weights_path = download_huggingface_weights(
        base_url=base_url, sub="vae", prefer_fp16=False
    )
    logger.debug(f"vae: {vae_weights_path}")
    vae_weights = translator.load_and_translate_weights(
        source_path=vae_weights_path,
        device="cpu",
    )
    lda = SDXLAutoencoderSliced(device="cpu", dtype=dtype)
    lda.load_state_dict(vae_weights, assign=True)
    del vae_weights

    translator = translators.diffusers_unet_sdxl_to_refiners_translator()
    unet_weights_path = download_huggingface_weights(
        base_url=base_url, sub="unet", prefer_fp16=True
    )
    logger.debug(f"unet: {unet_weights_path}")
    unet_weights = translator.load_and_translate_weights(
        source_path=unet_weights_path,
        device="cpu",
    )
    if for_inpainting:
        unet = SDXLUNet(device="cpu", dtype=dtype, in_channels=9)
    else:
        unet = SDXLUNet(device="cpu", dtype=dtype, in_channels=4)
    unet.load_state_dict(unet_weights, assign=True)
    del unet_weights

    text_encoder_1_path = download_huggingface_weights(
        base_url=base_url, sub="text_encoder"
    )
    text_encoder_2_path = download_huggingface_weights(
        base_url=base_url, sub="text_encoder_2"
    )
    logger.debug(f"text encoder 1: {text_encoder_1_path}")
    logger.debug(f"text encoder 2: {text_encoder_2_path}")
    text_encoder_weights = (
        translators.DoubleTextEncoderTranslator().load_and_translate_weights(
            text_encoder_l_weights_path=text_encoder_1_path,
            text_encoder_g_weights_path=text_encoder_2_path,
            device="cpu",
        )
    )
    text_encoder = DoubleTextEncoder(device="cpu", dtype=torch.float32)
    text_encoder.load_state_dict(text_encoder_weights, assign=True)
    del text_encoder_weights
    lda = lda.to(device=device, dtype=torch.float32)
    unet = unet.to(device=device, dtype=dtype)
    text_encoder = text_encoder.to(device=device, dtype=dtype)
    if for_inpainting:
        StableDiffusionCls = StableDiffusion_XL_Inpainting
    else:
        StableDiffusionCls = StableDiffusion_XL

    sd = StableDiffusionCls(
        device=device, dtype=None, lda=lda, unet=unet, clip_text_encoder=text_encoder
    )

    return sd


def load_sdxl_pipeline_from_compvis_weights(
    base_url: str, for_inpainting=False, device=None, dtype=torch.float16
):
    from imaginairy.utils import get_device

    device = device or get_device()
    unet_weights, vae_weights, text_encoder_weights = load_sdxl_compvis_weights(
        base_url
    )
    lda = SDXLAutoencoderSliced(device="cpu", dtype=dtype)
    lda.load_state_dict(vae_weights, assign=True)
    del vae_weights

    if for_inpainting:
        unet = SDXLUNet(device="cpu", dtype=dtype, in_channels=9)
    else:
        unet = SDXLUNet(device="cpu", dtype=dtype, in_channels=4)
    unet.load_state_dict(unet_weights, assign=True)
    del unet_weights

    text_encoder = DoubleTextEncoder(device="cpu", dtype=torch.float32)
    text_encoder.load_state_dict(text_encoder_weights, assign=True)
    del text_encoder_weights
    lda = lda.to(device=device, dtype=torch.float32)
    unet = unet.to(device=device)
    text_encoder = text_encoder.to(device=device)

    if for_inpainting:
        StableDiffusionCls = StableDiffusion_XL_Inpainting
    else:
        StableDiffusionCls = StableDiffusion_XL
    sd = StableDiffusionCls(
        device=device, dtype=None, lda=lda, unet=unet, clip_text_encoder=text_encoder
    )

    return sd


def load_sdxl_pipeline(base_url, device=None, for_inpainting=False):
    logger.info(f"Loading SDXL weights from {base_url}")
    device = device or get_device()

    with logger.timed_info(f"Loaded SDXL pipeline from {base_url}"):
        if is_diffusers_repo_url(base_url):
            sd = load_sdxl_pipeline_from_diffusers_weights(
                base_url, for_inpainting=for_inpainting, device=device
            )
        else:
            sd = load_sdxl_pipeline_from_compvis_weights(
                base_url, for_inpainting=for_inpainting, device=device
            )
        return sd


def open_weights(filepath, device=None):
    from imaginairy.utils import get_device

    if device is None:
        device = get_device()

    if "safetensor" in filepath.lower():
        from imaginairy.vendored.refiners.fluxion.utils import safe_open

        with safe_open(path=filepath, framework="pytorch", device=device) as tensors:
            state_dict = {
                key: tensors.get_tensor(key)
                for key in tensors.keys()  # noqa
            }
    else:
        import torch

        state_dict = torch.load(filepath, map_location=device)

    while "state_dict" in state_dict:
        state_dict = state_dict["state_dict"]

    return state_dict


def load_tensors(tensorfile, map_location=None):
    if tensorfile == "empty":
        # used for testing
        return {}
    if tensorfile.endswith((".ckpt", ".pth", ".bin")):
        return torch.load(tensorfile, map_location=map_location)
    if tensorfile.endswith(".safetensors"):
        return load_file(tensorfile, device=map_location)

    return load_file(tensorfile, device=map_location)

    # raise ValueError(f"Unknown tensorfile type: {tensorfile}")


def load_stable_diffusion_compvis_weights(weights_url):
    from imaginairy.utils import get_device
    from imaginairy.weight_management.conversion import cast_weights
    from imaginairy.weight_management.utils import (
        COMPONENT_NAMES,
        FORMAT_NAMES,
        MODEL_NAMES,
    )

    weights_path = get_cached_url_path(weights_url, category="weights")
    logger.info(f"Loading weights from {weights_path}")
    state_dict = open_weights(weights_path, device=get_device())

    text_encoder_prefix = "cond_stage_model."
    cut_start = len(text_encoder_prefix)
    text_encoder_state_dict = {
        k[cut_start:]: v
        for k, v in state_dict.items()
        if k.startswith(text_encoder_prefix)
    }
    text_encoder_state_dict = cast_weights(
        source_weights=text_encoder_state_dict,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.TEXT_ENCODER,
        source_format=FORMAT_NAMES.COMPVIS,
        dest_format=FORMAT_NAMES.DIFFUSERS,
    )
    text_encoder_state_dict = cast_weights(
        source_weights=text_encoder_state_dict,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.TEXT_ENCODER,
        source_format=FORMAT_NAMES.DIFFUSERS,
        dest_format=FORMAT_NAMES.REFINERS,
    )

    vae_prefix = "first_stage_model."
    cut_start = len(vae_prefix)
    vae_state_dict = {
        k[cut_start:]: v for k, v in state_dict.items() if k.startswith(vae_prefix)
    }
    vae_state_dict = cast_weights(
        source_weights=vae_state_dict,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.VAE,
        source_format=FORMAT_NAMES.COMPVIS,
        dest_format=FORMAT_NAMES.DIFFUSERS,
    )
    vae_state_dict = cast_weights(
        source_weights=vae_state_dict,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.VAE,
        source_format=FORMAT_NAMES.DIFFUSERS,
        dest_format=FORMAT_NAMES.REFINERS,
    )

    unet_prefix = "model."
    cut_start = len(unet_prefix)
    unet_state_dict = {
        k[cut_start:]: v for k, v in state_dict.items() if k.startswith(unet_prefix)
    }
    unet_state_dict = cast_weights(
        source_weights=unet_state_dict,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.UNET,
        source_format=FORMAT_NAMES.COMPVIS,
        dest_format=FORMAT_NAMES.DIFFUSERS,
    )

    unet_state_dict = cast_weights(
        source_weights=unet_state_dict,
        source_model_name=MODEL_NAMES.SD15,
        source_component_name=COMPONENT_NAMES.UNET,
        source_format=FORMAT_NAMES.DIFFUSERS,
        dest_format=FORMAT_NAMES.REFINERS,
    )

    return vae_state_dict, unet_state_dict, text_encoder_state_dict


def load_sdxl_compvis_weights(url):
    from safetensors import safe_open

    weights_path = get_cached_url_path(url)
    state_dict = {}
    unet_state_dict = {}
    vae_state_dict = {}
    text_encoder_1_state_dict = {}
    text_encoder_2_state_dict = {}
    with safe_open(weights_path, framework="pt") as f:
        for key in f.keys():  # noqa
            if key.startswith("model.diffusion_model."):
                unet_state_dict[key] = f.get_tensor(key)
            elif key.startswith("first_stage_model"):
                vae_state_dict[key] = f.get_tensor(key)
            elif key.startswith("conditioner.embedders.0."):
                text_encoder_1_state_dict[key] = f.get_tensor(key)
            elif key.startswith("conditioner.embedders.1."):
                text_encoder_2_state_dict[key] = f.get_tensor(key)
            else:
                state_dict[key] = f.get_tensor(key)
                logger.warning(f"Unused key {key}")

    unet_weightmap = load_weight_map("Compvis-UNet-SDXL-to-Diffusers")
    vae_weightmap = load_weight_map("Compvis-Autoencoder-SDXL-to-Diffusers")
    text_encoder_1_weightmap = load_weight_map("Compvis-TextEncoder-SDXL-to-Diffusers")
    text_encoder_2_weightmap = load_weight_map(
        "Compvis-OpenClipTextEncoder-SDXL-to-Diffusers"
    )

    diffusers_unet_state_dict = unet_weightmap.translate_weights(unet_state_dict)

    refiners_unet_state_dict = (
        diffusers_unet_sdxl_to_refiners_translator().translate_weights(
            diffusers_unet_state_dict
        )
    )

    diffusers_vae_state_dict = vae_weightmap.translate_weights(vae_state_dict)

    refiners_vae_state_dict = (
        diffusers_autoencoder_kl_to_refiners_translator().translate_weights(
            diffusers_vae_state_dict
        )
    )

    diffusers_text_encoder_1_state_dict = text_encoder_1_weightmap.translate_weights(
        text_encoder_1_state_dict
    )

    for key in list(text_encoder_2_state_dict.keys()):
        if key.endswith((".in_proj_bias", ".in_proj_weight")):
            value = text_encoder_2_state_dict[key]
            q, k, v = value.chunk(3, dim=0)
            text_encoder_2_state_dict[f"{key}.0"] = q
            text_encoder_2_state_dict[f"{key}.1"] = k
            text_encoder_2_state_dict[f"{key}.2"] = v
            del text_encoder_2_state_dict[key]

    diffusers_text_encoder_2_state_dict = text_encoder_2_weightmap.translate_weights(
        text_encoder_2_state_dict
    )

    refiners_text_encoder_weights = DoubleTextEncoderTranslator().translate_weights(
        diffusers_text_encoder_1_state_dict, diffusers_text_encoder_2_state_dict
    )
    return (
        refiners_unet_state_dict,
        refiners_vae_state_dict,
        refiners_text_encoder_weights,
    )



from typing import Callable

from torch import Tensor, device as Device, dtype as DType

import imaginairy.vendored.refiners.fluxion.layers as fl
from imaginairy.vendored.refiners.foundationals.clip.common import FeedForward, PositionalEncoder


class ClassToken(fl.Chain):
    def __init__(self, embedding_dim: int, device: Device | str | None = None, dtype: DType | None = None) -> None:
        self.embedding_dim = embedding_dim
        super().__init__(fl.Parameter(1, embedding_dim, device=device, dtype=dtype))


class PatchEncoder(fl.Chain):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        patch_size: int = 16,
        use_bias: bool = True,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.patch_size = patch_size
        self.use_bias = use_bias
        super().__init__(
            fl.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                kernel_size=(self.patch_size, self.patch_size),
                stride=(self.patch_size, self.patch_size),
                use_bias=self.use_bias,
                device=device,
                dtype=dtype,
            ),
            fl.Permute(0, 2, 3, 1),
        )


class TransformerLayer(fl.Chain):
    def __init__(
        self,
        embedding_dim: int = 768,
        feedforward_dim: int = 3072,
        num_attention_heads: int = 12,
        layer_norm_eps: float = 1e-5,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.embedding_dim = embedding_dim
        self.feedforward_dim = feedforward_dim
        self.num_attention_heads = num_attention_heads
        self.layer_norm_eps = layer_norm_eps
        super().__init__(
            fl.Residual(
                fl.LayerNorm(normalized_shape=embedding_dim, eps=layer_norm_eps, device=device, dtype=dtype),
                fl.SelfAttention(
                    embedding_dim=embedding_dim, num_heads=num_attention_heads, device=device, dtype=dtype
                ),
            ),
            fl.Residual(
                fl.LayerNorm(normalized_shape=embedding_dim, eps=layer_norm_eps, device=device, dtype=dtype),
                FeedForward(embedding_dim=embedding_dim, feedforward_dim=feedforward_dim, device=device, dtype=dtype),
            ),
        )


class ViTEmbeddings(fl.Chain):
    def __init__(
        self,
        image_size: int = 224,
        embedding_dim: int = 768,
        patch_size: int = 32,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.image_size = image_size
        self.embedding_dim = embedding_dim
        self.patch_size = patch_size
        super().__init__(
            fl.Concatenate(
                ClassToken(embedding_dim, device=device, dtype=dtype),
                fl.Chain(
                    PatchEncoder(
                        in_channels=3,
                        out_channels=embedding_dim,
                        patch_size=patch_size,
                        use_bias=False,
                        device=device,
                        dtype=dtype,
                    ),
                    fl.Reshape((image_size // patch_size) ** 2, embedding_dim),
                ),
                dim=1,
            ),
            fl.Residual(
                PositionalEncoder(
                    max_sequence_length=(image_size // patch_size) ** 2 + 1,
                    embedding_dim=embedding_dim,
                    device=device,
                    dtype=dtype,
                ),
            ),
        )


class CLIPImageEncoder(fl.Chain):
    def __init__(
        self,
        image_size: int = 224,
        embedding_dim: int = 768,
        output_dim: int = 512,
        patch_size: int = 32,
        num_layers: int = 12,
        num_attention_heads: int = 12,
        feedforward_dim: int = 3072,
        layer_norm_eps: float = 1e-5,
        device: Device | str | None = None,
        dtype: DType | None = None,
    ) -> None:
        self.image_size = image_size
        self.embedding_dim = embedding_dim
        self.output_dim = output_dim
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.num_attention_heads = num_attention_heads
        self.feedforward_dim = feedforward_dim
        cls_token_pooling: Callable[[Tensor], Tensor] = lambda x: x[:, 0, :]
        super().__init__(
            ViTEmbeddings(
                image_size=image_size, embedding_dim=embedding_dim, patch_size=patch_size, device=device, dtype=dtype
            ),
            fl.LayerNorm(normalized_shape=embedding_dim, eps=layer_norm_eps, device=device, dtype=dtype),
            fl.Chain(
                TransformerLayer(
                    embedding_dim=embedding_dim,
                    feedforward_dim=feedforward_dim,
                    num_attention_heads=num_attention_heads,
                    layer_norm_eps=layer_norm_eps,
                    device=device,
                    dtype=dtype,
                )
                for _ in range(num_layers)
            ),
            fl.Lambda(func=cls_token_pooling),
            fl.LayerNorm(normalized_shape=embedding_dim, eps=layer_norm_eps, device=device, dtype=dtype),
            fl.Linear(in_features=embedding_dim, out_features=output_dim, bias=False, device=device, dtype=dtype),
        )


class CLIPImageEncoderH(CLIPImageEncoder):
    def __init__(self, device: Device | str | None = None, dtype: DType | None = None) -> None:
        super().__init__(
            embedding_dim=1280,
            output_dim=1024,
            patch_size=14,
            num_layers=32,
            num_attention_heads=16,
            feedforward_dim=5120,
            device=device,
            dtype=dtype,
        )


class CLIPImageEncoderG(CLIPImageEncoder):
    def __init__(self, device: Device | str | None = None, dtype: DType | None = None) -> None:
        super().__init__(
            embedding_dim=1664,
            output_dim=1280,
            patch_size=14,
            num_layers=48,
            num_attention_heads=16,
            feedforward_dim=8192,
            device=device,
            dtype=dtype,
        )



import os
from functools import lru_cache

from torch import device as Device

from imaginairy.weight_management.translation import TensorDict, WeightTranslationMap

_current_folder = os.path.dirname(os.path.abspath(__file__))
weight_map_folder = os.path.join(_current_folder, "weight_maps")


@lru_cache
def load_weight_map(map_name: str) -> WeightTranslationMap:
    map_path = os.path.join(weight_map_folder, f"{map_name}.weightmap.json")
    return WeightTranslationMap.load(map_path)


def transformers_text_encoder_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Transformers-ClipTextEncoder")


def transformers_image_encoder_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Transformers-ClipImageEncoder-SD21")


def diffusers_autoencoder_kl_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-AutoencoderKL-SD")


def diffusers_unet_sd15_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-UNet-SD15")


def diffusers_unet_sdxl_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-UNet-SDXL")


def informative_drawings_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("InformativeDrawings")


def diffusers_controlnet_sd15_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-Controlnet-SD15")


def diffusers_ip_adapter_sd15_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-IPAdapter-SD15")


def diffusers_ip_adapter_sdxl_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-IPAdapter-SDXL")


def diffusers_ip_adapter_plus_sd15_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-IPAdapterPlus-SD15")


def diffusers_ip_adapter_plus_sdxl_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-IPAdapterPlus-SDXL")


def diffusers_t2i_adapter_sd15_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-T2IAdapter-SD15")


def diffusers_t2i_adapter_sdxl_to_refiners_translator() -> WeightTranslationMap:
    return load_weight_map("Diffusers-T2IAdapter-SDXL")


class DoubleTextEncoderTranslator:
    def __init__(self):
        self.translator = transformers_text_encoder_to_refiners_translator()

    def load_untranslated_weights(
        self,
        text_encoder_l_weights_path: str,
        text_encoder_g_weights_path: str,
        device: Device | str = "cpu",
    ) -> tuple[TensorDict, TensorDict]:
        text_encoder_l_weights = self.translator.load_untranslated_weights(
            text_encoder_l_weights_path, device=device
        )
        text_encoder_g_weights = self.translator.load_untranslated_weights(
            text_encoder_g_weights_path, device=device
        )
        return text_encoder_l_weights, text_encoder_g_weights

    def load_and_translate_weights(
        self,
        text_encoder_l_weights_path: str,
        text_encoder_g_weights_path: str,
        device: Device | str = "cpu",
    ) -> TensorDict:
        text_encoder_l_weights, text_encoder_g_weights = self.load_untranslated_weights(
            text_encoder_l_weights_path, text_encoder_g_weights_path, device=device
        )

        return self.translate_weights(text_encoder_l_weights, text_encoder_g_weights)

    def translate_weights(
        self, text_encoder_l_weights: TensorDict, text_encoder_g_weights: TensorDict
    ) -> TensorDict:
        new_sd: TensorDict = {}

        text_encoder_l_weights = self.translator.translate_weights(
            text_encoder_l_weights
        )
        text_encoder_g_weights = self.translator.translate_weights(
            text_encoder_g_weights
        )

        for k in list(text_encoder_l_weights.keys()):
            if k.startswith("TransformerLayer_12"):
                text_encoder_l_weights.pop(k)
            elif k.startswith("LayerNorm"):
                text_encoder_l_weights.pop(k)
            else:
                new_key = f"Parallel.CLIPTextEncoderL.{k}"
                new_sd[new_key] = text_encoder_l_weights.pop(k)

        new_sd["Parallel.TextEncoderWithPooling.Parallel.Chain.Linear.weight"] = (
            text_encoder_g_weights.pop("Linear.weight")
        )
        for k in list(text_encoder_g_weights.keys()):
            if k.startswith("TransformerLayer_32"):
                new_key = f"Parallel.TextEncoderWithPooling.Parallel.Chain.CLIPTextEncoderG.TransformerLayer{k[19:]}"
            elif k.startswith("LayerNorm"):
                new_key = f"Parallel.TextEncoderWithPooling.Parallel.Chain.CLIPTextEncoderG.{k}"
            else:
                new_key = f"Parallel.TextEncoderWithPooling.CLIPTextEncoderG.{k}"

            new_sd[new_key] = text_encoder_g_weights.pop(k)

        return new_sd



#
# import torch
#
# from imaginairy.utils import get_device
#
# torch.manual_seed(0)
# from transformers import CLIPTextModel, CLIPTokenizer
# from diffusers import AutoencoderKL, UNet2DConditionModel, LMSDiscreteScheduler
#
# from tqdm.auto import tqdm, trange
# from torch import autocast
# import PIL.Image as PImage
# from PIL import Image
# import numpy
# from torchvision import transforms
# import torchvision.transforms.functional as f
# import random
# import requests
# from io import BytesIO
#
# # import clip
# import open_clip as clip
# from torch import nn
# import torch.nn.functional as F
# import io
#
# offload_device = "cpu"
# model_name = "CompVis/stable-diffusion-v1-4"
# attention_slicing = True #@param {"type":"boolean"}
# unet_path = False
#
# vae = AutoencoderKL.from_pretrained(model_name, subfolder="vae", use_auth_token=True)
#
# tokenizer = CLIPTokenizer.from_pretrained("openai/clip-vit-large-patch14")
# try:
#     text_encoder = CLIPTextModel.from_pretrained(model_name, subfolder="text_encoder", use_auth_token=True)
# except:
#     print("Text encoder could not be loaded from the repo specified for some reason, falling back to the vit-l repo")
#     text_encoder = CLIPTextModel.from_pretrained("openai/clip-vit-large-patch14")
#
# if unet_path!=None:
#     # unet = UNet2DConditionModel.from_pretrained(unet_path)
#     from huggingface_hub import hf_hub_download
#     model_name = hf_hub_download(repo_id=unet_path, filename="unet.pt")
#     unet = torch.jit.load(model_name)
# else:
#     unet = UNet2DConditionModel.from_pretrained(model_name, subfolder="unet", use_auth_token=True)
#     if attention_slicing:
#         slice_size = unet.config.attention_head_dim // 2
#         unet.set_attention_slice(slice_size)
#
# scheduler = LMSDiscreteScheduler(beta_start=0.00085, beta_end=0.012, beta_schedule="scaled_linear", num_train_timesteps=1000)
#
# vae = vae.to(offload_device).half()
# text_encoder = text_encoder.to(offload_device).half()
# unet = unet.to(get_device()).half()
# class MakeCutouts(nn.Module):
#     def __init__(self, cut_size, cutn, cut_pow=1.):
#         super().__init__()
#         self.cut_size = cut_size
#         self.cutn = cutn
#         self.cut_pow = cut_pow
#
#     def forward(self, input):
#         sideY, sideX = input.shape[2:4]
#         max_size = min(sideX, sideY)
#         min_size = min(sideX, sideY, self.cut_size)
#         cutouts = []
#         for _ in range(self.cutn):
#             size = int(torch.rand([]) ** self.cut_pow * (max_size - min_size) + min_size)
#             offsetx = torch.randint(0, sideX - size + 1, ())
#             offsety = torch.randint(0, sideY - size + 1, ())
#             cutout = input[:, :, offsety:offsety + size, offsetx:offsetx + size]
#             cutouts.append(F.adaptive_avg_pool2d(cutout, self.cut_size))
#         return torch.cat(cutouts)
#
#
# to_tensor_tfm = transforms.ToTensor()
#
#
# # mismatch of tons of image encoding / decoding / loading functions i can't be asked to clean up right now
#
# def pil_to_latent(input_im):
#     # Single image -> single latent in a batch (so size 1, 4, 64, 64)
#     with torch.no_grad():
#         with autocast("cuda"):
#             latent = vae.encode(to_tensor_tfm(input_im.convert("RGB")).unsqueeze(0).to(
#                 get_device()) * 2 - 1).latent_dist  # Note scaling
#     #   print(latent)
#     return 0.18215 * latent.mode()  # or .mean or .sample
#
#
# def latents_to_pil(latents):
#     # bath of latents -> list of images
#     latents = (1 / 0.18215) * latents
#     with torch.no_grad():
#         image = vae.decode(latents)
#     image = (image / 2 + 0.5).clamp(0, 1)
#     image = image.detach().cpu().permute(0, 2, 3, 1).numpy()
#     images = (image * 255).round().astype("uint8")
#     pil_images = [Image.fromarray(image) for image in images]
#     return pil_images
#
#
# def get_latent_from_url(url, size=(512, 512)):
#     response = requests.get(url)
#     img = PImage.open(BytesIO(response.content))
#     img = img.resize(size).convert("RGB")
#     latent = pil_to_latent(img)
#     return latent
#
#
# def scale_and_decode(latents):
#     with autocast("cuda"):
#         # scale and decode the image latents with vae
#         latents = 1 / 0.18215 * latents
#         with torch.no_grad():
#             image = vae.decode(latents).sample.squeeze(0)
#         image = f.to_pil_image((image / 2 + 0.5).clamp(0, 1))
#         return image
#
#
# def fetch(url_or_path):
#     import io
#     if str(url_or_path).startswith('http://') or str(url_or_path).startswith('https://'):
#         r = requests.get(url_or_path)
#         r.raise_for_status()
#         fd = io.BytesIO()
#         fd.write(r.content)
#         fd.seek(0)
#         return PImage.open(fd).convert('RGB')
#     return PImage.open(open(url_or_path, 'rb')).convert('RGB')
#
#
# """
# grabs all text up to the first occurrence of ':'
# uses the grabbed text as a sub-prompt, and takes the value following ':' as weight
# if ':' has no value defined, defaults to 1.0
# repeats until no text remaining
# """
#
#
# def split_weighted_subprompts(text, split=":"):
#     remaining = len(text)
#     prompts = []
#     weights = []
#     while remaining > 0:
#         if split in text:
#             idx = text.index(split)  # first occurrence from start
#             # grab up to index as sub-prompt
#             prompt = text[:idx]
#             remaining -= idx
#             # remove from main text
#             text = text[idx + 1:]
#             # find value for weight
#             if " " in text:
#                 idx = text.index(" ")  # first occurrence
#             else:  # no space, read to end
#                 idx = len(text)
#             if idx != 0:
#                 try:
#                     weight = float(text[:idx])
#                 except:  # couldn't treat as float
#                     print(f"Warning: '{text[:idx]}' is not a value, are you missing a space?")
#                     weight = 1.0
#             else:  # no value found
#                 weight = 1.0
#             # remove from main text
#             remaining -= idx
#             text = text[idx + 1:]
#             # append the sub-prompt and its weight
#             prompts.append(prompt)
#             weights.append(weight)
#         else:  # no : found
#             if len(text) > 0:  # there is still text though
#                 # take remainder as weight 1
#                 prompts.append(text)
#                 weights.append(1.0)
#             remaining = 0
#     print(prompts, weights)
#     return prompts, weights
#
#
# # from some stackoverflow comment
# import numpy as np
#
#
# def lerp(a, b, x):
#     "linear interpolation"
#     return a + x * (b - a)
#
#
# def fade(t):
#     "6t^5 - 15t^4 + 10t^3"
#     return 6 * t ** 5 - 15 * t ** 4 + 10 * t ** 3
#
#
# def gradient(h, x, y):
#     "grad converts h to the right gradient vector and return the dot product with (x,y)"
#     vectors = np.array([[0, 1], [0, -1], [1, 0], [-1, 0]])
#     g = vectors[h % 4]
#     return g[:, :, 0] * x + g[:, :, 1] * y
#
#
# def perlin(x, y, seed=0):
#     # permutation table
#     np.random.seed(seed)
#     p = np.arange(256, dtype=int)
#     np.random.shuffle(p)
#     p = np.stack([p, p]).flatten()
#     # coordinates of the top-left
#     xi, yi = x.astype(int), y.astype(int)
#     # internal coordinates
#     xf, yf = x - xi, y - yi
#     # fade factors
#     u, v = fade(xf), fade(yf)
#     # noise components
#     n00 = gradient(p[p[xi] + yi], xf, yf)
#     n01 = gradient(p[p[xi] + yi + 1], xf, yf - 1)
#     n11 = gradient(p[p[xi + 1] + yi + 1], xf - 1, yf - 1)
#     n10 = gradient(p[p[xi + 1] + yi], xf - 1, yf)
#     # combine noises
#     x1 = lerp(n00, n10, u)
#     x2 = lerp(n01, n11, u)  # FIX1: I was using n10 instead of n01
#     return lerp(x1, x2, v)  # FIX2: I also had to reverse x1 and x2 here
#
#
# def sample(args):
#     global in_channels
#     global text_encoder  # uugghhhghhghgh
#     global vae  # UUGHGHHGHGH
#     global unet  # .hggfkgjks;ldjf
#     # prompt = args.prompt
#     prompts, weights = split_weighted_subprompts(args.prompt)
#     h, w = args.size
#     steps = args.steps
#     scale = args.scale
#     classifier_guidance = args.classifier_guidance
#     use_init = len(args.init_img) > 1
#     if args.seed != -1:
#         seed = args.seed
#         generator = torch.manual_seed(seed)
#     else:
#         seed = random.randint(0, 10_000)
#         generator = torch.manual_seed(seed)
#     print(f"Generating with seed {seed}...")
#
#     # tokenize / encode text
#     tokens = [tokenizer(prompt, padding="max_length", max_length=tokenizer.model_max_length, truncation=True,
#                         return_tensors="pt") for prompt in prompts]
#     with torch.no_grad():
#         # move CLIP to cuda
#         text_encoder = text_encoder.to(get_device())
#         text_embeddings = [text_encoder(tok.input_ids.to(get_device()))[0].unsqueeze(0) for tok in tokens]
#         text_embeddings = [text_embeddings[i] * weights[i] for i in range(len(text_embeddings))]
#         text_embeddings = torch.cat(text_embeddings, 0).sum(0)
#         max_length = 77
#         uncond_input = tokenizer(
#             [""], padding="max_length", max_length=max_length, return_tensors="pt"
#         )
#         uncond_embeddings = text_encoder(uncond_input.input_ids.to(get_device()))[0]
#         text_embeddings = torch.cat([uncond_embeddings, text_embeddings])
#         # move it back to CPU so there's more vram for generating
#         text_encoder = text_encoder.to(offload_device)
#     images = []
#
#     if args.lpips_guidance:
#         import lpips
#         lpips_model = lpips.LPIPS(net='vgg').to(get_device())
#         init = to_tensor_tfm(fetch(args.init_img).resize(args.size)).to(get_device())
#
#     for batch_n in trange(args.batches):
#         with autocast("cuda"):
#             # unet = unet.to(get_device())
#             scheduler.set_timesteps(steps)
#             if not use_init or args.start_step == 0:
#                 latents = torch.randn(
#                     (1, in_channels, h // 8, w // 8),
#                     generator=generator
#                 )
#                 latents = latents.to(get_device())
#                 latents = latents * scheduler.sigmas[0]
#                 start_step = args.start_step
#             else:
#                 # Start step
#                 start_step = args.start_step - 1
#                 start_sigma = scheduler.sigmas[start_step]
#                 start_timestep = int(scheduler.timesteps[start_step])
#
#                 # Prep latents
#                 vae = vae.to(get_device())
#                 encoded = get_latent_from_url(args.init_img)
#                 if not classifier_guidance:
#                     vae = vae.to(offload_device)
#
#                 noise = torch.randn_like(encoded)
#                 sigmas = scheduler.match_shape(scheduler.sigmas[start_step], noise)
#                 noisy_samples = encoded + noise * sigmas
#
#                 latents = noisy_samples.to(get_device()).half()
#
#             if args.perlin_multi != 0:
#                 linx = np.linspace(0, 5, h // 8, endpoint=False)
#                 liny = np.linspace(0, 5, w // 8, endpoint=False)
#                 x, y = np.meshgrid(liny, linx)
#                 p = [np.expand_dims(perlin(x, y, seed=i), 0) for i in range(4)]  # reproducible seed
#                 p = np.concatenate(p, 0)
#                 p = torch.tensor(p).unsqueeze(0).cuda()
#                 latents = latents + (p * args.perlin_multi).to(get_device()).half()
#
#             for i, t in tqdm(enumerate(scheduler.timesteps), total=steps):
#                 if i > start_step:
#                     latent_model_input = torch.cat([latents] * 2)
#                     sigma = scheduler.sigmas[i]
#                     latent_model_input = latent_model_input / ((sigma ** 2 + 1) ** 0.5)
#
#                     with torch.no_grad():
#                         # noise_pred = unet(latent_model_input, t, encoder_hidden_states=text_embeddings)["sample"]
#                         # noise_pred = unet(latent_model_input, torch.tensor(t, dtype=torch.float32).cuda().half(), text_embeddings)#["sample"]
#                         noise_pred = unet(latent_model_input, torch.tensor(t, dtype=torch.float32).cuda(),
#                                           text_embeddings)  # ["sample"]
#
#                     # cfg
#                     noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
#                     noise_pred = noise_pred_uncond + scale * (noise_pred_text - noise_pred_uncond)
#
#                     # cg
#                     if classifier_guidance:
#                         # vae = vae.to(get_device())
#                         if vae.device != latents.device:
#                             vae = vae.to(latents.device)
#                         latents = latents.detach().requires_grad_()
#                         latents_x0 = latents - sigma * noise_pred
#                         denoised_images = vae.decode((1 / 0.18215) * latents_x0).sample / 2 + 0.5
#                         if args.loss_scale != 0:
#                             loss = args.loss_fn(denoised_images) * args.loss_scale
#                         else:
#                             loss = 0
#                             init_losses = lpips_model(denoised_images, init)
#                             loss = loss + init_losses.sum() * args.lpips_scale
#
#                         cond_grad = -torch.autograd.grad(loss, latents)[0]
#                         latents = latents.detach() + cond_grad * sigma ** 2
#                         # vae = vae.to(offload_device)
#
#                     latents = scheduler.step(noise_pred, i, latents)["prev_sample"]
#         vae = vae.to(get_device())
#         output_image = scale_and_decode(latents)
#         vae = vae.to(offload_device)
#         images.append(output_image)
#
#         import gc
#         gc.collect()
#         torch.cuda.empty_cache()
#
#         images[-1].save(f"output/{batch_n}.png")
#
#     return images
#
# def test_guided_image():
#     prompt = "tardigrade portrait [intricate] [artstation]"  # @param {"type":"string"}
#
#     # prompt = add_suffixes(prompt)
#
#     init_img = ""  # @param {"type":"string"}
#     size = [640, 640]  # @param
#     steps = 65  # @param
#     start_step = 0  # @param
#     perlin_multi = 0.4  # @param
#     scale = 7  # @param
#     seed = -1  # @param
#     batches = 4  # @param
#     # @markdown ---
#
#     # @markdown ### Classifier Guidance
#     # @markdown `classifier_guidance` is whether or not to use the loss function in the previous cell to guide the image (slows down image generation a lot) <br>
#     # @markdown it also is very hit-and-miss in terms of quality, but can be really really good, try setting batches high and then taking a nap <br>
#     # @markdown `lpips_guidance` is for if you're using an init_img, it'll let you start closer to the beginning while trying to keep the overall shapes similar
#     # @markdown `lpips_scale` is similar to `loss_scale` but it's how much to push the model to keep the shapes the same <br>
#     # @markdown `loss_scale` is how much to guide according to that loss function <br>
#     # @markdown `clip_text_prompt` is a prompt for CLIP to optimize towards, if using classifier guidance (supports weighting with `prompt:weight`) <br>
#     # @markdown `clip_image_prompt` is an image url for CLIP to optimize towards if using classifier guidance (supports weighting with `url|weight` because of colons coming up in urls) <br>
#     # @markdown for `clip_model_name` and `clip_model_pretrained` check out the openclip repository https://github.com/mlfoundations/open_clip <br>
#     # @markdown `cutn` is the amount of permutations of the image to show to clip (can help with stability) <br>
#     # @markdown `accumulate` is how many times to run the image through the clip model, can help if you can only fit low cutn on the machine <br>
#     # @markdown *you cannot use the textual inversion tokens with the clip text prompt*  <br>
#     # @markdown *also clip guidance sucks for most things except removing very small details that dont make sense*
#     classifier_guidance = True  # @param {"type":"boolean"}
#     lpips_guidance = False  # @param {"type":"boolean"}
#     lpips_scale = 0  # @param
#     loss_scale = 1.  # @param
#
#     class BlankClass():
#         def __init__(self):
#             bruh = 'BRUH'
#
#     args = BlankClass()
#     args.prompt = prompt
#     args.init_img = init_img
#     args.size = size
#     args.steps = steps
#     args.start_step = start_step
#     args.scale = scale
#     args.perlin_multi = perlin_multi
#     args.seed = seed
#     args.batches = batches
#     args.classifier_guidance = classifier_guidance
#     args.lpips_guidance = lpips_guidance
#     args.lpips_scale = lpips_scale
#     args.loss_scale = loss_scale
#
#     loss_scale = 1
#     # make_cutouts = MakeCutouts(224, 16)
#
#     clip_text_prompt = "tardigrade portrait [intricate] [artstation]"  # @param {"type":"string"}
#     # clip_text_prompt = add_suffixes(clip_text_prompt)
#     clip_image_prompt = ""  # @param {"type":"string"}
#
#     if loss_scale != 0:
#         # clip_model = clip.load("ViT-B/32", jit=False)[0].eval().requires_grad_(False).to(get_device())
#         clip_model_name = "ViT-B-32"  # @param {"type":"string"}
#         clip_model_pretrained = "laion2b_s34b_b79k"  # @param {"type":"string"}
#         clip_model, _, preprocess = clip.create_model_and_transforms(clip_model_name, pretrained=clip_model_pretrained)
#         clip_model = clip_model.eval().requires_grad_(False).to(get_device())
#
#         cutn = 4  # @param
#         make_cutouts = MakeCutouts(clip_model.visual.image_size if type(clip_model.visual.image_size) != tuple else
#                                    clip_model.visual.image_size[0], cutn)
#
#     target = None
#     if len(clip_text_prompt) > 1:
#         clip_text_prompt, clip_text_weights = split_weighted_subprompts(clip_text_prompt)
#         target = clip_model.encode_text(clip.tokenize(clip_text_prompt).to(get_device())) * torch.tensor(
#             clip_text_weights).view(len(clip_text_prompt), 1).to(get_device())
#     if len(clip_image_prompt) > 1:
#         clip_image_prompt, clip_image_weights = split_weighted_subprompts(clip_image_prompt, split="|")
#         # pesky spaces
#         clip_image_prompt = [p.replace(" ", "") for p in clip_image_prompt]
#         images = [fetch(image) for image in clip_image_prompt]
#         images = [f.to_tensor(i).unsqueeze(0) for i in images]
#         images = [make_cutouts(i) for i in images]
#         encodings = [clip_model.encode_image(i.to(get_device())).mean(0) for i in images]
#
#         for i in range(len(encodings)):
#             encodings[i] = (encodings[i] * clip_image_weights[i]).unsqueeze(0)
#         # print(encodings.shape)
#         encodings = torch.cat(encodings, 0)
#         encoding = encodings.sum(0)
#
#         if target != None:
#             target = target + encoding
#         else:
#             target = encoding
#         target = target.half().to(get_device())
#
#     # free a little memory, we dont use the text encoder after this so just delete it
#     clip_model.transformer = None
#     import gc
#     gc.collect()
#     torch.cuda.empty_cache()
#
#     def spherical_distance(x, y):
#         x = F.normalize(x, dim=-1)
#         y = F.normalize(y, dim=-1)
#         l = (x - y).norm(dim=-1).div(2).arcsin().pow(2).mul(2).mean()
#         return l
#
#     def loss_fn(x):
#         with torch.autocast("cuda"):
#             cutouts = make_cutouts(x)
#             encoding = clip_model.encode_image(cutouts.float()).half()
#             loss = spherical_distance(encoding, target)
#             return loss.mean()
#
#     args.loss_fn = loss_fn
#
#
#     dtype = torch.float16
#     with torch.amp.autocast(device_type=get_device(), dtype=dtype):
#         output = sample(args)
#     print("Done!")



import numpy as np
from filterpy.kalman import KalmanFilter


def convert_bbox_to_z(bbox):
    """Takes a bounding box in the form [x1,y1,x2,y2] and returns z in the form
    [x,y,s,r] where x,y is the centre of the box and s is the scale/area and
    r is the aspect ratio
    """
    w = bbox[2] - bbox[0]
    h = bbox[3] - bbox[1]
    x = bbox[0] + w / 2.
    y = bbox[1] + h / 2.
    s = w * h  # scale is just area
    r = w / float(h)
    return np.array([x, y, s, r]).reshape((4, 1))


def convert_x_to_bbox(x, score=None):
    """Takes a bounding box in the centre form [x,y,s,r] and returns it in
    the form [x1,y1,x2,y2] where x1,y1 is the top left and x2,y2 is the bottom
    right
    """
    w = np.sqrt(x[2] * x[3])
    h = x[2] / w
    if score is None:
        return np.array([x[0] - w / 2., x[1] - h / 2., x[0] + w / 2., x[1] + h / 2.]).reshape((1, 4))
    else:
        return np.array([x[0] - w / 2., x[1] - h / 2., x[0] + w / 2., x[1] + h / 2., score]).reshape((1, 5))


class KalmanBoxTracker(object):
    """This class represents the internal state of individual tracked objects
    observed as bbox.
    doc: https://filterpy.readthedocs.io/en/latest/kalman/KalmanFilter.html
    """
    count = 0

    def __init__(self, bbox):
        """Initialize a tracker using initial bounding box.
        """
        # define constant velocity model
        # TODO: x: what is the meanning of x[4:7], v?
        self.kf = KalmanFilter(dim_x=7, dim_z=4)
        # F (dim_x, dim_x): state transition matrix
        self.kf.F = np.array([[1, 0, 0, 0, 1, 0, 0], [0, 1, 0, 0, 0, 1, 0], [0, 0, 1, 0, 0, 0,
                                                                             1], [0, 0, 0, 1, 0, 0, 0],
                              [0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 1]])
        # H (dim_z, dim_x): measurement function
        self.kf.H = np.array([[1, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0],
                              [0, 0, 0, 1, 0, 0, 0]])
        # R (dim_z, dim_z): measurement uncertainty/noise
        self.kf.R[2:, 2:] *= 10.
        # P (dim_x, dim_x): covariance matrix
        # give high uncertainty to the unobservable initial velocities
        self.kf.P[4:, 4:] *= 1000.
        self.kf.P *= 10.
        # Q (dim_x, dim_x): Process uncertainty/noise
        self.kf.Q[-1, -1] *= 0.01
        self.kf.Q[4:, 4:] *= 0.01
        # x (dim_x, 1): filter state estimate
        self.kf.x[:4] = convert_bbox_to_z(bbox)

        self.time_since_update = 0
        self.id = KalmanBoxTracker.count
        KalmanBoxTracker.count += 1
        self.history = []
        self.hits = 0
        self.hit_streak = 0
        self.age = 0

        # 解决画面中无人脸检测到时而导致的原有追踪器人像预测的漂移bug
        self.predict_num = 0  # 连续预测的数目

        # additional fields
        self.face_attributes = []

    def update(self, bbox):
        """Updates the state vector with observed bbox.
        """
        self.time_since_update = 0
        self.history = []
        self.hits += 1
        self.hit_streak += 1  # 连续命中
        if bbox != []:
            self.kf.update(convert_bbox_to_z(bbox))
            self.predict_num = 0
        else:
            self.predict_num += 1

    def predict(self):
        """Advances the state vector and returns the predicted bounding box
        estimate.
        """

        if (self.kf.x[6] + self.kf.x[2]) <= 0:
            self.kf.x[6] *= 0.0
        self.kf.predict()
        self.age += 1
        if self.time_since_update > 0:
            self.hit_streak = 0
        self.time_since_update += 1
        self.history.append(convert_x_to_bbox(self.kf.x))
        return self.history[-1][0]

    def get_state(self):
        """Returns the current bounding box estimate."""
        return convert_x_to_bbox(self.kf.x)[0]



from abc import ABC, abstractmethod
from typing import TypeVar

import torch
from PIL import Image
from torch import Tensor, device as Device, dtype as DType

import imaginairy.vendored.refiners.fluxion.layers as fl
from imaginairy.vendored.refiners.foundationals.latent_diffusion.auto_encoder import LatentDiffusionAutoencoder
from imaginairy.vendored.refiners.foundationals.latent_diffusion.schedulers.scheduler import Scheduler

T = TypeVar("T", bound="fl.Module")

TLatentDiffusionModel = TypeVar("TLatentDiffusionModel", bound="LatentDiffusionModel")


class LatentDiffusionModel(fl.Module, ABC):
    def __init__(
        self,
        unet: fl.Module,
        lda: LatentDiffusionAutoencoder,
        clip_text_encoder: fl.Module,
        scheduler: Scheduler,
        device: Device | str = "cpu",
        dtype: DType = torch.float32,
    ) -> None:
        super().__init__()
        self.device: Device = device if isinstance(device, Device) else Device(device=device)
        self.dtype = dtype
        self.unet = unet.to(device=self.device, dtype=self.dtype)
        self.lda = lda.to(device=self.device, dtype=self.dtype)
        self.clip_text_encoder = clip_text_encoder.to(device=self.device, dtype=self.dtype)
        self.scheduler = scheduler.to(device=self.device, dtype=self.dtype)

    def set_inference_steps(self, num_steps: int, first_step: int = 0) -> None:
        initial_diffusion_rate = self.scheduler.initial_diffusion_rate
        final_diffusion_rate = self.scheduler.final_diffusion_rate
        device, dtype = self.scheduler.device, self.scheduler.dtype
        self.scheduler = self.scheduler.__class__(
            num_inference_steps=num_steps,
            initial_diffusion_rate=initial_diffusion_rate,
            final_diffusion_rate=final_diffusion_rate,
            first_inference_step=first_step,
        ).to(device=device, dtype=dtype)

    def init_latents(
        self,
        size: tuple[int, int],
        init_image: Image.Image | None = None,
        noise: Tensor | None = None,
    ) -> Tensor:
        height, width = size
        if noise is None:
            noise = torch.randn(1, 4, height // 8, width // 8, device=self.device)
        assert list(noise.shape[2:]) == [
            height // 8,
            width // 8,
        ], f"noise shape is not compatible: {noise.shape}, with size: {size}"
        if init_image is None:
            return noise
        encoded_image = self.lda.encode_image(image=init_image.resize(size=(width, height)))
        return self.scheduler.add_noise(
            x=encoded_image,
            noise=noise,
            step=self.scheduler.first_inference_step,
        )

    @property
    def steps(self) -> list[int]:
        return self.scheduler.inference_steps

    @abstractmethod
    def set_unet_context(self, *, timestep: Tensor, clip_text_embedding: Tensor, **_: Tensor) -> None:
        ...

    @abstractmethod
    def set_self_attention_guidance(self, enable: bool, scale: float = 1.0) -> None:
        ...

    @abstractmethod
    def has_self_attention_guidance(self) -> bool:
        ...

    @abstractmethod
    def compute_self_attention_guidance(
        self, x: Tensor, noise: Tensor, step: int, *, clip_text_embedding: Tensor, **kwargs: Tensor
    ) -> Tensor:
        ...

    def forward(
        self, x: Tensor, step: int, *, clip_text_embedding: Tensor, condition_scale: float = 7.5, **kwargs: Tensor
    ) -> Tensor:
        timestep = self.scheduler.timesteps[step].unsqueeze(dim=0)
        self.set_unet_context(timestep=timestep, clip_text_embedding=clip_text_embedding, **kwargs)

        latents = torch.cat(tensors=(x, x))  # for classifier-free guidance
        # scale latents for schedulers that need it
        latents = self.scheduler.scale_model_input(latents, step=step)
        unconditional_prediction, conditional_prediction = self.unet(latents).chunk(2)

        # classifier-free guidance
        noise = unconditional_prediction + condition_scale * (conditional_prediction - unconditional_prediction)
        x = x.narrow(dim=1, start=0, length=4)  # support > 4 channels for inpainting

        if self.has_self_attention_guidance():
            noise += self.compute_self_attention_guidance(
                x=x, noise=unconditional_prediction, step=step, clip_text_embedding=clip_text_embedding, **kwargs
            )

        return self.scheduler(x, noise=noise, step=step)

    def structural_copy(self: TLatentDiffusionModel) -> TLatentDiffusionModel:
        return self.__class__(
            unet=self.unet.structural_copy(),
            lda=self.lda.structural_copy(),
            clip_text_encoder=self.clip_text_encoder.structural_copy(),
            scheduler=self.scheduler,
            device=self.device,
            dtype=self.dtype,
        )



"""Functions for managing prompt schedules"""

import csv
import re
from copy import copy

from imaginairy.schema import ImaginePrompt
from imaginairy.utils import frange


def parse_schedule_str(schedule_str):
    """Parse a schedule string into a list of values."""
    pattern = re.compile(r"([a-zA-Z0-9_-]+)\[([a-zA-Z0-9_:,. -]+)\]")
    match = pattern.match(schedule_str)
    if not match:
        msg = f"Invalid kwarg schedule: {schedule_str}"
        raise ValueError(msg)

    arg_name = match.group(1).replace("-", "_")
    if not hasattr(ImaginePrompt(), arg_name):
        msg = f"Invalid kwarg schedule. Not a valid argument name: {arg_name}"
        raise ValueError(msg)

    arg_values = match.group(2)
    if ":" in arg_values:
        start, end, step = arg_values.split(":")
        arg_values = list(frange(float(start), float(end), float(step)))
    else:
        arg_values = parse_csv_line(arg_values)
    return arg_name, arg_values


def parse_schedule_strs(schedule_strs):
    """Parse and validate input prompt schedules."""
    schedules = {}
    for schedule_str in schedule_strs:
        arg_name, arg_values = parse_schedule_str(schedule_str)
        schedules[arg_name] = arg_values

    # Validate that all schedules have the same length
    schedule_lengths = [len(v) for v in schedules.values()]
    if len(set(schedule_lengths)) > 1:
        raise ValueError("All schedules must have the same length")

    return schedules


def prompt_mutator(prompt, schedules):
    """
    Given a prompt and a list of kwarg schedules, return a series of prompts that follow the schedule.

    kwarg_schedules example:
    {
        "prompt_strength": [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
    }

    """
    schedule_length = len(next(iter(schedules.values())))
    for i in range(schedule_length):
        new_prompt = copy(prompt)
        for attr_name, schedule in schedules.items():
            setattr(new_prompt, attr_name, schedule[i])
        new_prompt.validate()
        yield new_prompt


def parse_csv_line(line):
    reader = csv.reader([line])
    for row in reader:
        parsed_row = []
        for value in row:
            try:
                parsed_row.append(float(value))
            except ValueError:
                parsed_row.append(value)
        return parsed_row



"""Functions for image upscaling using RealESRGAN"""

import numpy as np
import torch
from PIL import Image

from imaginairy.utils import get_device
from imaginairy.utils.downloads import get_cached_url_path
from imaginairy.utils.model_cache import memory_managed_model
from imaginairy.vendored.basicsr.rrdbnet_arch import RRDBNet
from imaginairy.vendored.realesrgan import RealESRGANer


@memory_managed_model("realesrgan_upsampler", memory_usage_mb=70)
def realesrgan_upsampler(tile=512, tile_pad=50, ultrasharp=False):
    model = RRDBNet(
        num_in_ch=3, num_out_ch=3, num_feat=64, num_block=23, num_grow_ch=32, scale=4
    )
    if ultrasharp:
        url = "https://huggingface.co/lokCX/4x-Ultrasharp/resolve/1856559b50de25116a7c07261177dd128f1f5664/4x-UltraSharp.pth"
    else:
        url = "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.1.0/RealESRGAN_x4plus.pth"

    model_path = get_cached_url_path(url)
    device = get_device()

    upsampler = RealESRGANer(
        scale=4,
        model_path=model_path,
        model=model,
        tile=tile,
        device=device,
        tile_pad=tile_pad,
        half=True,
    )

    upsampler.device = torch.device(device)
    upsampler.model.to(device)

    return upsampler


def upscale_image(img, ultrasharp=False):
    img = img.convert("RGB")

    np_img = np.array(img, dtype=np.uint8)
    upsampler_output, img_mode = realesrgan_upsampler(ultrasharp=ultrasharp).enhance(
        np_img[:, :, ::-1]
    )
    return Image.fromarray(upsampler_output[:, :, ::-1], mode=img_mode)
